#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import param_utils
from heat.tests import common


class TestExtractBool(common.HeatTestCase):
    def test_extract_bool(self):
        for value in ('True', 'true', 'TRUE', True):
            self.assertTrue(param_utils.extract_bool('bool', value))
        for value in ('False', 'false', 'FALSE', False):
            self.assertFalse(param_utils.extract_bool('bool', value))
        for value in ('foo', 't', 'f', 'yes', 'no', 'y', 'n', '1', '0', None):
            self.assertRaises(ValueError, param_utils.extract_bool,
                              'bool', value)


class TestExtractInt(common.HeatTestCase):
    def test_extract_int(self):
        # None case
        self.assertIsNone(param_utils.extract_int('num', None))

        # 0 case
        self.assertEqual(0, param_utils.extract_int('num', 0))
        self.assertEqual(0, param_utils.extract_int('num', 0, allow_zero=True))
        self.assertEqual(0, param_utils.extract_int('num', '0'))
        self.assertEqual(0, param_utils.extract_int('num', '0',
                                                    allow_zero=True))
        self.assertRaises(ValueError,
                          param_utils.extract_int,
                          'num', 0, allow_zero=False)
        self.assertRaises(ValueError,
                          param_utils.extract_int,
                          'num', '0', allow_zero=False)

        # positive values
        self.assertEqual(1, param_utils.extract_int('num', 1))
        self.assertEqual(1, param_utils.extract_int('num', '1'))
        self.assertRaises(ValueError, param_utils.extract_int, 'num', '1.1')
        self.assertRaises(ValueError, param_utils.extract_int, 'num', 1.1)

        # negative values
        self.assertEqual(-1, param_utils.extract_int('num', -1,
                                                     allow_negative=True))
        self.assertEqual(-1, param_utils.extract_int('num', '-1',
                                                     allow_negative=True))
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', '-1.1',
                          allow_negative=True)
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', -1.1,
                          allow_negative=True)

        self.assertRaises(ValueError, param_utils.extract_int, 'num', -1)
        self.assertRaises(ValueError, param_utils.extract_int, 'num', '-1')
        self.assertRaises(ValueError, param_utils.extract_int, 'num', '-1.1')
        self.assertRaises(ValueError, param_utils.extract_int, 'num', -1.1)

        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', -1,
                          allow_negative=False)
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', '-1',
                          allow_negative=False)
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', '-1.1',
                          allow_negative=False)
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', -1.1,
                          allow_negative=False)

        # Non-int value
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', 'abc')
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', '')
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', 'true')
        self.assertRaises(ValueError,
                          param_utils.extract_int, 'num', True)


class TestExtractTags(common.HeatTestCase):
    def test_extract_tags(self):
        self.assertRaises(ValueError, param_utils.extract_tags, "aaaaaaaaaaaaa"
                          "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                          "aaaaaaaaaaaaaaaaa,a")
        self.assertEqual(["foo", "bar"], param_utils.extract_tags('foo,bar'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import sys

from cryptography import fernet
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes
from cryptography.hazmat.primitives import padding
from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils import encodeutils

from heat.common import exception
from heat.common.i18n import _

auth_opts = [
    cfg.StrOpt('auth_encryption_key',
               secret=True,
               default='notgood but just long enough i t',
               help=_('Key used to encrypt authentication info in the '
                      'database. Length of this key must be 32 characters.'))
]

cfg.CONF.register_opts(auth_opts)


class SymmetricCrypto(object):
    """Symmetric Key Crypto object.

    This class creates a Symmetric Key Crypto object that can be used
    to decrypt arbitrary data.

    Note: This is a reimplementation of the decryption algorithm from
    oslo-incubator, and is provided for backward compatibility. Once we have a
    DB migration script available to re-encrypt using new encryption method as
    part of upgrade, this can be removed.

    :param enctype: Encryption Cipher name (default: AES)
    """

    def __init__(self, enctype='AES'):
        self.algo = algorithms.AES

    def decrypt(self, key, msg, b64decode=True):
        """Decrypts the provided ciphertext.

        The ciphertext can be optionally base64 encoded.

        Uses AES-128-CBC with an IV by default.

        :param key: The Encryption key.
        :param msg: the ciphetext, the first block is the IV

        :returns: the plaintext message, after padding is removed.
        """
        key = str.encode(get_valid_encryption_key(key))
        if b64decode:
            msg = base64.b64decode(msg)
        algo = self.algo(key)
        block_size_bytes = algo.block_size // 8
        iv = msg[:block_size_bytes]
        backend = backends.default_backend()
        cipher = Cipher(algo, modes.CBC(iv), backend=backend)
        decryptor = cipher.decryptor()
        padded = (decryptor.update(msg[block_size_bytes:]) +
                  decryptor.finalize())
        unpadder = padding.ANSIX923(algo.block_size).unpadder()
        plain = unpadder.update(padded) + unpadder.finalize()
        # The original padding algorithm was a slight variation on ANSI X.923,
        # where the size of the padding did not include the byte that tells
        # you the size of the padding. Therefore, we need to remove one extra
        # byte (which will be 0x00) when unpadding.
        return plain[:-1]


def encrypt(value, encryption_key=None):
    if value is None:
        return None, None
    encryption_key = get_valid_encryption_key(encryption_key, fix_length=True)
    encoded_key = base64.b64encode(encryption_key.encode('utf-8'))
    sym = fernet.Fernet(encoded_key)
    res = sym.encrypt(encodeutils.safe_encode(value))
    return 'cryptography_decrypt_v1', encodeutils.safe_decode(res)


def decrypt(method, data, encryption_key=None):
    if method is None or data is None:
        return None
    decryptor = getattr(sys.modules[__name__], method)
    value = decryptor(data, encryption_key)
    if value is not None:
        return encodeutils.safe_decode(value, 'utf-8')


def encrypted_dict(data, encryption_key=None):
    'Return an encrypted dict. Values converted to json before encrypted'
    return_data = {}
    if not data:
        return return_data
    for prop_name, prop_value in data.items():
        prop_string = jsonutils.dumps(prop_value)
        encrypted_value = encrypt(prop_string, encryption_key)
        return_data[prop_name] = encrypted_value
    return return_data


def decrypted_dict(data, encryption_key=None):
    'Return a decrypted dict. Assume input values are encrypted json fields.'
    return_data = {}
    if not data:
        return return_data
    for prop_name, prop_value in data.items():
        method, value = prop_value
        try:
            decrypted_value = decrypt(method, value, encryption_key)
        except UnicodeDecodeError:
            # The dict contained valid JSON on the way in, so if what comes
            # out is garbage then the key was incorrect.
            raise exception.InvalidEncryptionKey()
        prop_string = jsonutils.loads(decrypted_value)
        return_data[prop_name] = prop_string
    return return_data


def oslo_decrypt_v1(value, encryption_key=None):
    encryption_key = get_valid_encryption_key(encryption_key)
    sym = SymmetricCrypto()
    return sym.decrypt(encryption_key, value, b64decode=True)


def cryptography_decrypt_v1(value, encryption_key=None):
    encryption_key = get_valid_encryption_key(encryption_key, fix_length=True)
    encoded_key = base64.b64encode(encryption_key.encode('utf-8'))
    sym = fernet.Fernet(encoded_key)
    try:
        return sym.decrypt(encodeutils.safe_encode(value))
    except fernet.InvalidToken:
        raise exception.InvalidEncryptionKey()


def get_valid_encryption_key(encryption_key, fix_length=False):
    if encryption_key is None:
        encryption_key = cfg.CONF.auth_encryption_key

    if fix_length and len(encryption_key) < 32:
        # Backward compatible size
        encryption_key = encryption_key * 2

    return encryption_key[:32]


def heat_decrypt(value, encryption_key=None):
    """Decrypt data that has been encrypted using an older version of Heat.

    Note: the encrypt function returns the function that is needed to
    decrypt the data. The database then stores this. When the data is
    then retrieved (potentially by a later version of Heat) the decrypt
    function must still exist. So whilst it may seem that this function
    is not referenced, it will be referenced from the database.
    """
    encryption_key = str.encode(get_valid_encryption_key(encryption_key))
    auth = base64.b64decode(value)
    AES = algorithms.AES(encryption_key)
    block_size_bytes = AES.block_size // 8
    iv = auth[:block_size_bytes]
    backend = backends.default_backend()
    cipher = Cipher(AES, modes.CFB(iv), backend=backend)
    decryptor = cipher.decryptor()
    return decryptor.update(auth[block_size_bytes:]) + decryptor.finalize()


def list_opts():
    yield None, auth_opts



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from monascaclient import exc as monasca_exc
from monascaclient.v2_0 import client as monasca_client

from heat.common import exception as heat_exc
from heat.engine.clients import client_plugin
from heat.engine import constraints

CLIENT_NAME = 'monasca'


class MonascaClientPlugin(client_plugin.ClientPlugin):
    exceptions_module = [monasca_exc]

    service_types = [MONITORING] = ['monitoring']

    VERSION = '2_0'

    def _create(self):
        interface = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        endpoint = self.url_for(service_type=self.MONITORING,
                                endpoint_type=interface)

        # Directly use v2_0 client to avoid dynamic import in monasca client,
        # We can switch back once https://review.opendev.org/#/c/700989 fixed.
        return monasca_client.Client(
            session=self.context.keystone_session,
            service_type='monitoring',
            endpoint=endpoint)

    def is_not_found(self, ex):
        return isinstance(ex, monasca_exc.NotFound)

    def is_un_processable(self, ex):
        return isinstance(ex, monasca_exc.UnprocessableEntity)

    def get_notification(self, notification):
        try:
            return self.client().notifications.get(
                notification_id=notification)['id']
        except monasca_exc.NotFound:
            raise heat_exc.EntityNotFound(entity='Monasca Notification',
                                          name=notification)


class MonascaNotificationConstraint(constraints.BaseCustomConstraint):

    resource_client_name = CLIENT_NAME
    resource_getter_name = 'get_notification'



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.octavia import octavia_base
from heat.engine import support
from heat.engine import translation


class PoolMember(octavia_base.OctaviaBase):
    """A resource for managing Octavia Pool Members.

    A pool member represents a single backend node.
    """

    PROPERTIES = (
        POOL, ADDRESS, PROTOCOL_PORT, MONITOR_ADDRESS, MONITOR_PORT,
        WEIGHT, ADMIN_STATE_UP, SUBNET, TAGS
    ) = (
        'pool', 'address', 'protocol_port', 'monitor_address', 'monitor_port',
        'weight', 'admin_state_up', 'subnet', 'tags'
    )

    ATTRIBUTES = (
        ADDRESS_ATTR, POOL_ID_ATTR
    ) = (
        'address', 'pool_id'
    )

    properties_schema = {
        POOL: properties.Schema(
            properties.Schema.STRING,
            _('Name or ID of the load balancing pool.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('octavia.pool')
            ]
        ),
        ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('IP address of the pool member on the pool network.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
        PROTOCOL_PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('Port on which the pool member listens for requests or '
              'connections.'),
            required=True,
            constraints=[
                constraints.Range(1, 65535),
            ]
        ),
        MONITOR_ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('Alternate IP address which health monitor can use for '
              'health check.'),
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
        MONITOR_PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('Alternate Port which health monitor can use for health check.'),
            constraints=[
                constraints.Range(1, 65535),
            ]
        ),
        WEIGHT: properties.Schema(
            properties.Schema.INTEGER,
            _('Weight of pool member in the pool (default to 1).'),
            default=1,
            constraints=[
                constraints.Range(0, 256),
            ],
            update_allowed=True
        ),
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of the pool member.'),
            default=True,
            update_allowed=True
        ),
        SUBNET: properties.Schema(
            properties.Schema.STRING,
            _('Subnet name or ID of this member.'),
            constraints=[
                constraints.CustomConstraint('neutron.subnet')
            ],
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('A list of simple strings assigned to the member. The property '
              'is supported with Stein Octavia or newer version.'),
            update_allowed=True,
            support_status=support.SupportStatus(version='13.0.0'),
        ),
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.SUBNET],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='subnet'
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.POOL],
                client_plugin=self.client_plugin(),
                finder='get_pool'
            ),
        ]

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items() if v is not None)
        props.pop(self.POOL)
        if self.SUBNET in props:
            props['subnet_id'] = props.pop(self.SUBNET)
        return props

    def _resource_create(self, properties):
        pool = self.properties[self.POOL]
        return self.client().member_create(
            pool, json={'member': properties})['member']

    def _resource_update(self, prop_diff):
        pool = self.properties[self.POOL]
        self.client().member_set(pool,
                                 self.resource_id,
                                 json={'member': prop_diff})

    def _resource_delete(self):
        pool = self.properties[self.POOL]
        if pool:
            self.client().member_delete(pool, self.resource_id)

    def _show_resource(self):
        pool = self.properties[self.POOL]
        return self.client().member_show(pool, self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::PoolMember': PoolMember,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.common import exceptions as qe
from neutronclient.neutron import v2_0 as neutronV20
from neutronclient.v2_0 import client as neutronclient
from oslo_serialization import jsonutils

from heat.common import exception
from heat.common import template_format
from heat.engine import resource
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


neutron_port_template = '''
heat_template_version: 2015-04-30
description: Template to test port Neutron resource
resources:
  port:
    type: OS::Neutron::Port
    properties:
      network: net1234
      fixed_ips:
        - subnet: sub1234
          ip_address: 10.0.3.21
      device_owner: network:dhcp
'''


neutron_port_with_address_pair_template = '''
heat_template_version: 2015-04-30
description: Template to test port Neutron resource
resources:
  port:
    type: OS::Neutron::Port
    properties:
      network: abcd1234
      allowed_address_pairs:
        - ip_address: 10.0.3.21/8
          mac_address: 00-B0-D0-86-BB-F7
'''


neutron_port_with_no_fixed_ips_template = '''
heat_template_version: 2015-04-30
description: Template to test port Neutron resource
resources:
  port:
    type: OS::Neutron::Port
    properties:
      network: abcd1234
      no_fixed_ips: true
'''

neutron_port_security_template = '''
heat_template_version: 2015-04-30
description: Template to test port Neutron resource
resources:
  port:
    type: OS::Neutron::Port
    properties:
      network: abcd1234
      port_security_enabled: False
'''


neutron_port_propagate_ul_status_template = '''
heat_template_version: 2015-04-30
description: Template to test port Neutron resource
resources:
  port:
    type: OS::Neutron::Port
    properties:
      network: abcd1234
      propagate_uplink_status: True
'''


class NeutronPortTest(common.HeatTestCase):

    def setUp(self):
        super(NeutronPortTest, self).setUp()
        self.create_mock = self.patchobject(
            neutronclient.Client, 'create_port')
        self.port_show_mock = self.patchobject(
            neutronclient.Client, 'show_port')
        self.update_mock = self.patchobject(
            neutronclient.Client, 'update_port')
        self.subnet_show_mock = self.patchobject(
            neutronclient.Client, 'show_subnet')
        self.network_show_mock = self.patchobject(
            neutronclient.Client, 'show_network')
        self.find_mock = self.patchobject(
            neutronV20, 'find_resourceid_by_name_or_id')

    def test_missing_network(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'] = {}
        stack = utils.parse_stack(t)
        port = stack['port']
        self.assertRaises(exception.StackValidationFailed, port.validate)

    def test_missing_subnet_id(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['fixed_ips'][0].pop('subnet')
        stack = utils.parse_stack(t)
        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {
            'port': {
                "status": "BUILD",
                "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}
        self.port_show_mock.return_value = {
            'port': {
                "status": "ACTIVE",
                "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net1234',
            'fixed_ips': [
                {'ip_address': '10.0.3.21'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }})
        self.port_show_mock.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766')

    def test_missing_ip_address(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['fixed_ips'][0].pop('ip_address')
        stack = utils.parse_stack(t)
        self.find_mock.return_value = 'net_or_sub'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net_or_sub',
            'fixed_ips': [
                {'subnet_id': 'net_or_sub'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }})
        self.port_show_mock.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766')

    def test_missing_fixed_ips(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'].pop('fixed_ips')
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
            "fixed_ips": {
                "subnet_id": "d0e971a6-a6b4-4f4c-8c88-b75e9c120b7e",
                "ip_address": "10.0.0.2"
            }
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net1234',
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }})

    def test_allowed_address_pair(self):
        t = template_format.parse(neutron_port_with_address_pair_template)
        t['resources']['port']['properties'][
            'allowed_address_pairs'][0]['ip_address'] = '10.0.3.21'
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'allowed_address_pairs': [{
                'ip_address': '10.0.3.21',
                'mac_address': '00-B0-D0-86-BB-F7'
            }],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_id': '',
            'device_owner': ''
        }})

    def test_no_fixed_ips(self):
        t = template_format.parse(neutron_port_with_no_fixed_ips_template)
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'

        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}

        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'name': utils.PhysName(stack.name, 'port'),
            'fixed_ips': [],
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_id': '',
            'device_owner': ''
        }})

    def test_port_security_enabled(self):
        t = template_format.parse(neutron_port_security_template)
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'

        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}

        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'port_security_enabled': False,
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_id': '',
            'device_owner': ''
            }})

    def test_port_propagate_uplink_status(self):
        t = template_format.parse(neutron_port_propagate_ul_status_template)
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'

        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}

        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'propagate_uplink_status': True,
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_id': '',
            'device_owner': ''
            }})

    def test_missing_mac_address(self):
        t = template_format.parse(neutron_port_with_address_pair_template)
        t['resources']['port']['properties']['allowed_address_pairs'][0].pop(
            'mac_address'
        )
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'allowed_address_pairs': [{
                'ip_address': '10.0.3.21/8',
            }],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_owner': '',
            'device_id': ''}})

    def test_ip_address_is_cidr(self):
        t = template_format.parse(neutron_port_with_address_pair_template)
        t['resources']['port']['properties'][
            'allowed_address_pairs'][0]['ip_address'] = '10.0.3.0/24'
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'abcd1234'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "2e00180a-ff9d-42c4-b701-a0606b243447"
        }}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "2e00180a-ff9d-42c4-b701-a0606b243447"
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'abcd1234',
            'allowed_address_pairs': [{
                'ip_address': '10.0.3.0/24',
                'mac_address': '00-B0-D0-86-BB-F7'
            }],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'binding:vnic_type': 'normal',
            'device_owner': '',
            'device_id': ''
        }})

    def _mock_create_with_props(self):
        self.find_mock.return_value = 'net_or_sub'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"}}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
            "dns_assignment": {
                "hostname": "my-vm",
                "ip_address": "10.0.0.15",
                "fqdn": "my-vm.openstack.org."}

        }}

    def test_create_with_tags(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['tags'] = ['tag1', 'tag2']
        stack = utils.parse_stack(t)

        port_prop = {
            'network_id': 'net_or_sub',
            'fixed_ips': [
                {'subnet_id': 'net_or_sub', 'ip_address': '10.0.3.21'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }

        set_tag_mock = self.patchobject(neutronclient.Client, 'replace_tag')
        self._mock_create_with_props()

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': port_prop})
        set_tag_mock.assert_called_with('ports', port.resource_id,
                                        {'tags': ['tag1', 'tag2']})

    def test_security_groups(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['security_groups'] = [
            '8a2f582a-e1cd-480f-b85d-b02631c10656',
            '024613dc-b489-4478-b46f-ada462738740']
        stack = utils.parse_stack(t)

        port_prop = {
            'network_id': 'net_or_sub',
            'security_groups': ['8a2f582a-e1cd-480f-b85d-b02631c10656',
                                '024613dc-b489-4478-b46f-ada462738740'],
            'fixed_ips': [
                {'subnet_id': 'net_or_sub', 'ip_address': '10.0.3.21'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }

        self._mock_create_with_props()

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': port_prop})

    def test_port_with_dns_name(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['dns_name'] = 'myvm'
        stack = utils.parse_stack(t)

        port_prop = {
            'network_id': 'net_or_sub',
            'dns_name': 'myvm',
            'fixed_ips': [
                {'subnet_id': 'net_or_sub', 'ip_address': '10.0.3.21'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }

        self._mock_create_with_props()
        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual('my-vm.openstack.org.',
                         port.FnGetAtt('dns_assignment')['fqdn'])
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': port_prop})

    def test_security_groups_empty_list(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['security_groups'] = []
        stack = utils.parse_stack(t)

        port_prop = {
            'network_id': 'net_or_sub',
            'security_groups': [],
            'fixed_ips': [
                {'subnet_id': 'net_or_sub', 'ip_address': '10.0.3.21'}
            ],
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }

        self._mock_create_with_props()

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': port_prop})

    def test_update_failed_port_no_replace(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        port = stack['port']
        port.resource_id = 'r_id'
        port.state_set(port.CREATE, port.FAILED)
        new_props = port.properties.data.copy()
        new_props['name'] = 'new_one'
        self.find_mock.return_value = 'net_or_sub'
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
            "fixed_ips": {
                "subnet_id": "d0e971a6-a6b4-4f4c-8c88-b75e9c120b7e",
                "ip_address": "10.0.3.21"}}}
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)
        scheduler.TaskRunner(port.update, update_snippet)()
        self.assertEqual((port.UPDATE, port.COMPLETE), port.state)
        self.assertEqual(1, self.update_mock.call_count)

    def test_port_needs_update(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'].pop('fixed_ips')
        stack = utils.parse_stack(t)

        props = {'network_id': 'net1234',
                 'name': utils.PhysName(stack.name, 'port'),
                 'admin_state_up': True,
                 'device_owner': 'network:dhcp',
                 'device_id': '',
                 'binding:vnic_type': 'normal'}

        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {'port': {
            "status": "BUILD",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
            "fixed_ips": {
                "subnet_id": "d0e971a6-a6b4-4f4c-8c88-b75e9c120b7e",
                "ip_address": "10.0.0.2"
            }
        }}

        # create port
        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': props})

        new_props = props.copy()
        # test always replace
        new_props['replacement_policy'] = 'REPLACE_ALWAYS'
        new_props['network'] = new_props.pop('network_id')
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)
        self.assertRaises(resource.UpdateReplace, port._needs_update,
                          update_snippet, port.frozen_definition(),
                          new_props, port.properties, None)

        # test deferring to Resource._needs_update
        new_props['replacement_policy'] = 'AUTO'
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)
        self.assertTrue(port._needs_update(update_snippet,
                                           port.frozen_definition(),
                                           new_props, port.properties, None))

    def test_port_needs_update_network(self):
        net1 = '9cfe6c74-c105-4906-9a1f-81d9064e9bca'
        net2 = '0064eec9-5681-4ba7-a745-6f8e32db9503'
        props = {'network_id': net1,
                 'name': 'test_port',
                 'device_owner': 'network:dhcp',
                 'binding:vnic_type': 'normal',
                 'device_id': ''
                 }
        create_kwargs = props.copy()
        create_kwargs['admin_state_up'] = True

        self.find_mock.side_effect = [net1] * 8 + [net2] * 2 + [net1]
        self.create_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
        }}
        self.port_show_mock.return_value = {'port': {
            "status": "ACTIVE",
            "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
            "fixed_ips": {
                "subnet_id": "d0e971a6-a6b4-4f4c-8c88-b75e9c120b7e",
                "ip_address": "10.0.0.2"
            }
        }}

        # create port with network_id
        tmpl = neutron_port_template.replace(
            'network: net1234',
            'network_id: 9cfe6c74-c105-4906-9a1f-81d9064e9bca')
        t = template_format.parse(tmpl)
        t['resources']['port']['properties'].pop('fixed_ips')
        t['resources']['port']['properties']['name'] = 'test_port'
        stack = utils.parse_stack(t)

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertEqual((port.CREATE, port.COMPLETE), port.state)
        self.create_mock.assert_called_once_with({'port': create_kwargs})

        # Switch from network_id=ID to network=ID (no replace)
        new_props = props.copy()
        new_props['network'] = new_props.pop('network_id')
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)

        scheduler.TaskRunner(port.update, update_snippet)()
        self.assertEqual((port.UPDATE, port.COMPLETE), port.state)
        self.assertEqual(0, self.update_mock.call_count)

        # Switch from network=ID to network=NAME (no replace)
        new_props['network'] = 'net1234'
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)

        scheduler.TaskRunner(port.update, update_snippet)()
        self.assertEqual((port.UPDATE, port.COMPLETE), port.state)
        self.assertEqual(0, self.update_mock.call_count)

        # Switch to a different network (replace)
        new_props['network'] = 'net5678'
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      new_props)
        updater = scheduler.TaskRunner(port.update, update_snippet)
        self.assertRaises(resource.UpdateReplace, updater)
        self.assertEqual(11, self.find_mock.call_count)

    def test_get_port_attributes(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'].pop('fixed_ips')
        stack = utils.parse_stack(t)

        subnet_dict = {'name': 'test-subnet', 'enable_dhcp': True,
                       'network_id': 'net1234', 'dns_nameservers': [],
                       'tenant_id': '58a61fc3992944ce971404a2ece6ff98',
                       'ipv6_ra_mode': None, 'cidr': '10.0.0.0/24',
                       'allocation_pools': [{'start': '10.0.0.2',
                                             'end': '10.0.0.254'}],
                       'gateway_ip': '10.0.0.1', 'ipv6_address_mode': None,
                       'ip_version': 4, 'host_routes': [],
                       'id': 'd0e971a6-a6b4-4f4c-8c88-b75e9c120b7e'}
        network_dict = {'name': 'test-network', 'status': 'ACTIVE',
                        'router:external': False,
                        'availability_zone_hints': [],
                        'availability_zones': ['nova'],
                        'ipv4_address_scope': None, 'description': '',
                        'subnets': [subnet_dict['id']],
                        'port_security_enabled': True,
                        'propagate_uplink_status': True,
                        'tenant_id': '58a61fc3992944ce971404a2ece6ff98',
                        'tags': [], 'ipv6_address_scope': None,
                        'project_id': '58a61fc3992944ce971404a2ece6ff98',
                        'revision_number': 4, 'admin_state_up': True,
                        'shared': False, 'mtu': 1450, 'id': 'net1234'}
        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {'port': {
            'status': 'BUILD',
            'id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766'
        }}
        self.subnet_show_mock.return_value = {'subnet': subnet_dict}
        self.network_show_mock.return_value = {'network': network_dict}
        self.port_show_mock.return_value = {'port': {
            'status': 'DOWN',
            'name': utils.PhysName(stack.name, 'port'),
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'network_id': 'net1234',
            'device_id': 'dc68eg2c-b60g-4b3f-bd82-67ec87650532',
            'mac_address': 'fa:16:3e:75:67:60',
            'tenant_id': '58a61fc3992944ce971404a2ece6ff98',
            'security_groups': ['5b15d80c-6b70-4a1c-89c9-253538c5ade6'],
            'fixed_ips': [{'subnet_id': 'd0e971a6-a6b4-4f4c-8c88-b75e9c120b7e',
                           'ip_address': '10.0.0.2'}]
        }}

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net1234',
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''
        }})
        self.assertEqual('DOWN', port.FnGetAtt('status'))
        self.assertEqual([], port.FnGetAtt('allowed_address_pairs'))
        self.assertTrue(port.FnGetAtt('admin_state_up'))
        self.assertEqual('net1234', port.FnGetAtt('network_id'))
        self.assertEqual('fa:16:3e:75:67:60', port.FnGetAtt('mac_address'))
        self.assertEqual(utils.PhysName(stack.name, 'port'),
                         port.FnGetAtt('name'))
        self.assertEqual('dc68eg2c-b60g-4b3f-bd82-67ec87650532',
                         port.FnGetAtt('device_id'))
        self.assertEqual('58a61fc3992944ce971404a2ece6ff98',
                         port.FnGetAtt('tenant_id'))
        self.assertEqual(['5b15d80c-6b70-4a1c-89c9-253538c5ade6'],
                         port.FnGetAtt('security_groups'))
        self.assertEqual([{'subnet_id': 'd0e971a6-a6b4-4f4c-8c88-b75e9c120b7e',
                           'ip_address': '10.0.0.2'}],
                         port.FnGetAtt('fixed_ips'))
        self.assertEqual([subnet_dict], port.FnGetAtt('subnets'))
        self.assertEqual(network_dict, port.FnGetAtt('network'))
        self.assertRaises(exception.InvalidTemplateAttribute,
                          port.FnGetAtt, 'Foo')

    def test_subnet_attribute_exception(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'].pop('fixed_ips')
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {'port': {
            'status': 'BUILD',
            'id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766'
        }}
        self.port_show_mock.return_value = {'port': {
            'status': 'DOWN',
            'name': utils.PhysName(stack.name, 'port'),
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'network_id': 'net1234',
            'device_id': 'dc68eg2c-b60g-4b3f-bd82-67ec87650532',
            'mac_address': 'fa:16:3e:75:67:60',
            'tenant_id': '58a61fc3992944ce971404a2ece6ff98',
            'security_groups': ['5b15d80c-6b70-4a1c-89c9-253538c5ade6'],
            'fixed_ips': [{'subnet_id': 'd0e971a6-a6b4-4f4c-8c88-b75e9c120b7e',
                           'ip_address': '10.0.0.2'}]
        }}
        self.subnet_show_mock.side_effect = (qe.NeutronClientException(
            'ConnectionFailed: Connection to neutron failed: Maximum '
            'attempts reached'))

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertIsNone(port.FnGetAtt('subnets'))
        log_msg = ('Failed to fetch resource attributes: ConnectionFailed: '
                   'Connection to neutron failed: Maximum attempts reached')
        self.assertIn(log_msg, self.LOG.output)
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net1234',
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''}}
        )

    def test_network_attribute_exception(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties'].pop('fixed_ips')
        stack = utils.parse_stack(t)

        self.find_mock.return_value = 'net1234'
        self.create_mock.return_value = {'port': {
            'status': 'BUILD',
            'id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766'
        }}
        self.port_show_mock.return_value = {'port': {
            'status': 'DOWN',
            'name': utils.PhysName(stack.name, 'port'),
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'network_id': 'net1234',
            'device_id': 'dc68eg2c-b60g-4b3f-bd82-67ec87650532',
            'mac_address': 'fa:16:3e:75:67:60',
            'tenant_id': '58a61fc3992944ce971404a2ece6ff98',
            'security_groups': ['5b15d80c-6b70-4a1c-89c9-253538c5ade6'],
            'fixed_ips': [{'subnet_id': 'd0e971a6-a6b4-4f4c-8c88-b75e9c120b7e',
                           'ip_address': '10.0.0.2'}]
        }}
        self.network_show_mock.side_effect = (qe.NeutronClientException(
            'ConnectionFailed: Connection to neutron failed: Maximum '
            'attempts reached'))

        port = stack['port']
        scheduler.TaskRunner(port.create)()
        self.assertIsNone(port.FnGetAtt('network'))
        log_msg = ('Failed to fetch resource attributes: ConnectionFailed: '
                   'Connection to neutron failed: Maximum attempts reached')
        self.assertIn(log_msg, self.LOG.output)
        self.create_mock.assert_called_once_with({'port': {
            'network_id': 'net1234',
            'name': utils.PhysName(stack.name, 'port'),
            'admin_state_up': True,
            'device_owner': 'network:dhcp',
            'binding:vnic_type': 'normal',
            'device_id': ''}}
        )

    def test_prepare_for_replace_port_not_created(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        port = stack['port']
        port._show_resource = mock.Mock()
        port.data_set = mock.Mock()
        n_client = mock.Mock()
        port.client = mock.Mock(return_value=n_client)

        self.assertIsNone(port.resource_id)

        # execute prepare_for_replace
        port.prepare_for_replace()

        # check, if the port is not created, do nothing in
        # prepare_for_replace()
        self.assertFalse(port._show_resource.called)
        self.assertFalse(port.data_set.called)
        self.assertFalse(n_client.update_port.called)

    def test_prepare_for_replace_port_not_found(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        port = stack['port']
        port.resource_id = 'test_res_id'
        port._show_resource = mock.Mock(side_effect=qe.NotFound)
        port.data_set = mock.Mock()
        n_client = mock.Mock()
        port.client = mock.Mock(return_value=n_client)

        # execute prepare_for_replace
        port.prepare_for_replace()

        # check, if the port is not found, do nothing in
        # prepare_for_replace()
        self.assertTrue(port._show_resource.called)
        self.assertFalse(port.data_set.called)
        self.assertFalse(n_client.update_port.called)

    def test_prepare_for_replace_port(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        port = stack['port']
        port.resource_id = 'test_res_id'
        _value = {
            'fixed_ips': {
                'subnet_id': 'test_subnet',
                'ip_address': '42.42.42.42'
            }
        }
        port._show_resource = mock.Mock(return_value=_value)
        port.data_set = mock.Mock()
        n_client = mock.Mock()
        port.client = mock.Mock(return_value=n_client)

        # execute prepare_for_replace
        port.prepare_for_replace()

        # check, that data was stored
        port.data_set.assert_called_once_with(
            'port_fip', jsonutils.dumps(_value.get('fixed_ips')))

        # check, that port was updated and ip was removed
        expected_props = {'port': {'fixed_ips': []}}
        n_client.update_port.assert_called_once_with('test_res_id',
                                                     expected_props)

    def test_restore_prev_rsrc(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        new_port = stack['port']
        new_port.resource_id = 'new_res_id'
        # mock backup stack to return only one mocked old_port
        old_port = mock.Mock()
        new_port.stack._backup_stack = mock.Mock()
        new_port.stack._backup_stack().resources.get.return_value = old_port
        old_port.resource_id = 'old_res_id'
        _value = {
            'subnet_id': 'test_subnet',
            'ip_address': '42.42.42.42'
        }
        old_port.data = mock.Mock(
            return_value={'port_fip': jsonutils.dumps(_value)})

        n_client = mock.Mock()
        new_port.client = mock.Mock(return_value=n_client)

        # execute restore_prev_rsrc
        new_port.restore_prev_rsrc()

        # check, that ports were updated: old port get ip and
        # same ip was removed from old port
        expected_new_props = {'port': {'fixed_ips': []}}
        expected_old_props = {'port': {'fixed_ips': _value}}
        n_client.update_port.assert_has_calls([
            mock.call('new_res_id', expected_new_props),
            mock.call('old_res_id', expected_old_props)])

    def test_restore_prev_rsrc_convergence(self):
        t = template_format.parse(neutron_port_template)
        stack = utils.parse_stack(t)
        stack.store()

        # mock resource from previous template
        prev_rsrc = stack['port']
        prev_rsrc.resource_id = 'prev-rsrc'
        # store in db
        prev_rsrc.state_set(prev_rsrc.UPDATE, prev_rsrc.COMPLETE)

        # mock resource from existing template and store in db
        existing_rsrc = stack['port']
        existing_rsrc.current_template_id = stack.t.id
        existing_rsrc.resource_id = 'existing-rsrc'
        existing_rsrc.state_set(existing_rsrc.UPDATE, existing_rsrc.COMPLETE)

        # mock previous resource was replaced by existing resource
        prev_rsrc.replaced_by = existing_rsrc.id
        _value = {
            'subnet_id': 'test_subnet',
            'ip_address': '42.42.42.42'
        }
        prev_rsrc._data = {'port_fip': jsonutils.dumps(_value)}

        n_client = mock.Mock()
        prev_rsrc.client = mock.Mock(return_value=n_client)

        # execute restore_prev_rsrc
        prev_rsrc.restore_prev_rsrc(convergence=True)

        expected_existing_props = {'port': {'fixed_ips': []}}
        expected_prev_props = {'port': {'fixed_ips': _value}}
        n_client.update_port.assert_has_calls([
            mock.call(existing_rsrc.resource_id, expected_existing_props),
            mock.call(prev_rsrc.resource_id, expected_prev_props)])

    def test_port_get_live_state(self):
        t = template_format.parse(neutron_port_template)
        t['resources']['port']['properties']['value_specs'] = {
            'binding:vif_type': 'test'}

        stack = utils.parse_stack(t)

        port = stack['port']

        resp = {'port': {
            'status': 'DOWN',
            'binding:host_id': '',
            'name': 'flip-port-xjbal77qope3',
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'network_id': 'd6859535-efef-4184-b236-e5fcae856e0f',
            'dns_name': '',
            'extra_dhcp_opts': [],
            'mac_address': 'fa:16:3e:fe:64:79',
            'qos_policy_id': 'some',
            'dns_assignment': [],
            'binding:vif_details': {},
            'binding:vif_type': 'unbound',
            'device_owner': '',
            'tenant_id': '30f466e3d14b4251853899f9c26e2b66',
            'binding:profile': {},
            'port_security_enabled': True,
            'propagate_uplink_status': True,
            'binding:vnic_type': 'normal',
            'fixed_ips': [
                {'subnet_id': '02d9608f-8f30-4611-ad02-69855c82457f',
                 'ip_address': '10.0.3.4'}],
            'id': '829bf5c1-b59c-40ad-80e3-ea15a93879f3',
            'security_groups': ['c276247f-50fd-4289-862a-80fb81a55de1'],
            'device_id': ''}
        }
        port.client().show_port = mock.MagicMock(return_value=resp)
        port.resource_id = '1234'
        port._data = {}
        port.data_set = mock.Mock()

        reality = port.get_live_state(port.properties)
        expected = {
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'device_owner': '',
            'port_security_enabled': True,
            'propagate_uplink_status': True,
            'binding:vnic_type': 'normal',
            'fixed_ips': [
                {'subnet': '02d9608f-8f30-4611-ad02-69855c82457f',
                 'ip_address': '10.0.3.4'}],
            'security_groups': ['c276247f-50fd-4289-862a-80fb81a55de1'],
            'device_id': '',
            'dns_name': '',
            'qos_policy': 'some',
            'value_specs': {'binding:vif_type': 'unbound'}
        }

        self.assertEqual(set(expected.keys()), set(reality.keys()))
        for key in expected:
            self.assertEqual(expected[key], reality[key])


class UpdatePortTest(common.HeatTestCase):
    scenarios = [
        ('with_secgrp', dict(secgrp=['8a2f582a-e1cd-480f-b85d-b02631c10656'],
                             name='test',
                             value_specs={},
                             fixed_ips=None,
                             addr_pair=None,
                             vnic_type=None)),
        ('with_no_name', dict(secgrp=['8a2f582a-e1cd-480f-b85d-b02631c10656'],
                              orig_name='original',
                              name=None,
                              value_specs={},
                              fixed_ips=None,
                              addr_pair=None,
                              vnic_type=None)),
        ('with_empty_values', dict(secgrp=[],
                                   name='test',
                                   value_specs={},
                                   fixed_ips=[],
                                   addr_pair=[],
                                   vnic_type=None)),
        ('with_fixed_ips', dict(secgrp=None,
                                value_specs={},
                                fixed_ips=[
                                    {"subnet_id": "d0e971a6-a6b4-4f4c",
                                     "ip_address": "10.0.0.2"}],
                                addr_pair=None,
                                vnic_type=None)),
        ('with_addr_pair', dict(secgrp=None,
                                value_specs={},
                                fixed_ips=None,
                                addr_pair=[{'ip_address': '10.0.3.21',
                                            'mac_address': '00-B0-D0-86'}],
                                vnic_type=None)),

        ('with_value_specs', dict(secgrp=None,
                                  value_specs={'binding:vnic_type': 'direct'},
                                  fixed_ips=None,
                                  addr_pair=None,
                                  vnic_type=None)),
        ('normal_vnic', dict(secgrp=None,
                             value_specs={},
                             fixed_ips=None,
                             addr_pair=None,
                             vnic_type='normal')),
        ('direct_vnic', dict(secgrp=None,
                             value_specs={},
                             fixed_ips=None,
                             addr_pair=None,
                             vnic_type='direct')),
        ('physical_direct_vnic', dict(secgrp=None,
                                      value_specs={},
                                      fixed_ips=None,
                                      addr_pair=None,
                                      vnic_type='direct-physical')),
        ('baremetal_vnic', dict(secgrp=None,
                                value_specs={},
                                fixed_ips=None,
                                addr_pair=None,
                                vnic_type='baremetal')),
        ('virtio_forwarder_vnic', dict(secgrp=None,
                                       value_specs={},
                                       fixed_ips=None,
                                       addr_pair=None,
                                       vnic_type='virtio-forwarder')),
        ('smart_nic_vnic', dict(secgrp=None,
                                value_specs={},
                                fixed_ips=None,
                                addr_pair=None,
                                vnic_type='smart-nic')),
        ('with_all', dict(secgrp=['8a2f582a-e1cd-480f-b85d-b02631c10656'],
                          value_specs={},
                          fixed_ips=[
                              {"subnet_id": "d0e971a6-a6b4-4f4c",
                               "ip_address": "10.0.0.2"}],
                          addr_pair=[{'ip_address': '10.0.3.21',
                                      'mac_address': '00-B0-D0-86-BB-F7'}],
                          vnic_type='normal')),

        ]

    def test_update_port(self):
        t = template_format.parse(neutron_port_template)
        create_name = getattr(self, 'orig_name', None)
        if create_name is not None:
            t['resources']['port']['properties']['name'] = create_name
        stack = utils.parse_stack(t)

        def res_id(client, resource, name_or_id, cmd_resource=None):
            return {
                'network': 'net1234',
                'subnet': 'sub1234',
            }[resource]

        self.patchobject(neutronV20, 'find_resourceid_by_name_or_id',
                         side_effect=res_id)

        create_port_result = {
            'port': {
                "status": "BUILD",
                "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766"
            }
        }
        show_port_result = {
            'port': {
                "status": "ACTIVE",
                "id": "fc68ea2c-b60b-4b4f-bd82-94ec81110766",
                "fixed_ips": {
                    "subnet_id": "d0e971a6-a6b4-4f4c-8c88-b75e9c120b7e",
                    "ip_address": "10.0.0.2"
                }
            }
        }

        create_port = self.patchobject(neutronclient.Client, 'create_port',
                                       return_value=create_port_result)
        update_port = self.patchobject(neutronclient.Client, 'update_port')
        show_port = self.patchobject(neutronclient.Client, 'show_port',
                                     return_value=show_port_result)

        fake_groups_list = {
            'security_groups': [
                {
                    'tenant_id': 'dc4b074874244f7693dd65583733a758',
                    'id': '0389f747-7785-4757-b7bb-2ab07e4b09c3',
                    'name': 'default',
                    'security_group_rules': [],
                    'description': 'no protocol'
                }
            ]
        }
        self.patchobject(neutronclient.Client, 'list_security_groups',
                         return_value=fake_groups_list)
        set_tag_mock = self.patchobject(neutronclient.Client, 'replace_tag')

        props = {'network_id': 'net1234',
                 'fixed_ips': [{'subnet_id': 'sub1234',
                                'ip_address': '10.0.3.21'}],
                 'name': (create_name if create_name is not None else
                          utils.PhysName(stack.name, 'port')),
                 'admin_state_up': True,
                 'device_owner': 'network:dhcp',
                 'device_id': '',
                 'binding:vnic_type': 'normal'}

        update_props = props.copy()
        update_props['name'] = getattr(self, 'name', create_name)
        update_props['security_groups'] = self.secgrp
        update_props['value_specs'] = self.value_specs
        update_props['tags'] = ['test_tag']
        if self.fixed_ips:
            update_props['fixed_ips'] = self.fixed_ips
        update_props['allowed_address_pairs'] = self.addr_pair
        update_props['binding:vnic_type'] = self.vnic_type

        update_dict = update_props.copy()

        if update_props['allowed_address_pairs'] is None:
            update_dict['allowed_address_pairs'] = []

        if update_props['security_groups'] is None:
            update_dict['security_groups'] = [
                '0389f747-7785-4757-b7bb-2ab07e4b09c3',
            ]

        if update_props['name'] is None:
            update_dict['name'] = utils.PhysName(stack.name, 'port')

        value_specs = update_dict.pop('value_specs')
        if value_specs:
            for value_spec in value_specs.items():
                update_dict[value_spec[0]] = value_spec[1]

        tags = update_dict.pop('tags')

        # create port
        port = stack['port']
        self.assertIsNone(scheduler.TaskRunner(port.create)())
        create_port.assert_called_once_with({'port': props})
        # update port
        update_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      update_props)
        self.assertIsNone(scheduler.TaskRunner(port.handle_update,
                                               update_snippet, {},
                                               update_props)())

        update_port.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766', {'port': update_dict})
        set_tag_mock.assert_called_with('ports', port.resource_id,
                                        {'tags': tags})
        # check, that update does not cause of Update Replace
        create_snippet = rsrc_defn.ResourceDefinition(port.name, port.type(),
                                                      props)
        after_props, before_props = port._prepare_update_props(update_snippet,
                                                               create_snippet)
        self.assertIsNotNone(
            port.update_template_diff_properties(after_props, before_props))

        # With fixed_ips removed
        scheduler.TaskRunner(port.handle_update, update_snippet,
                             {}, {'fixed_ips': None})()

        # update with empty prop_diff
        scheduler.TaskRunner(port.handle_update, update_snippet, {}, {})()
        self.assertEqual(1, update_port.call_count)
        show_port.assert_called_with('fc68ea2c-b60b-4b4f-bd82-94ec81110766')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.engine.cfn import functions as cfn_funcs
from heat.engine.clients.os import monasca as client_plugin
from heat.engine.resources.openstack.monasca import notification
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils


sample_template = {
    'heat_template_version': '2015-10-15',
    'resources': {
        'test_resource': {
            'type': 'OS::Monasca::Notification',
            'properties': {
                'name': 'test-notification',
                'type': 'webhook',
                'address': 'http://localhost:80/',
                'period': 60
            }
        }
    }
}


class MonascaNotificationTest(common.HeatTestCase):

    def setUp(self):
        super(MonascaNotificationTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )

        self.test_resource = self.stack['test_resource']

        # Mock client
        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

        # Mock client plugin
        self.test_client_plugin = client_plugin.MonascaClientPlugin(self.ctx)
        self.test_client_plugin._create = mock.MagicMock(
            return_value=self.test_client)
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

    def _get_mock_resource(self):
        value = dict(id='477e8273-60a7-4c41-b683-fdb0bc7cd152')

        return value

    def test_validate_success_no_period(self):
        self.test_resource.properties.data.pop('period')
        self.test_resource.validate()

    def test_validate_invalid_type_with_period(self):
        self.test_resource.properties.data['type'] = self.test_resource.EMAIL
        self.assertRaises(exception.StackValidationFailed,
                          self.test_resource.validate)

    def test_validate_no_scheme_address_for_get_attr(self):
        self.test_resource.properties.data['type'] = self.test_resource.WEBHOOK
        self.patchobject(cfn_funcs, 'GetAtt', return_value=None)
        get_att = cfn_funcs.GetAtt(self.stack, 'Fn::GetAtt',
                                   ["ResourceA", "abc"])
        self.test_resource.properties.data['address'] = get_att
        self.assertIsNone(self.test_resource.validate())

    def test_validate_no_scheme_address_for_webhook(self):
        self.test_resource.properties.data['type'] = self.test_resource.WEBHOOK
        self.test_resource.properties.data['address'] = 'abc@def.com'
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Address "abc@def.com" doesn\'t have '
                         'required URL scheme', str(ex))

    def test_validate_no_netloc_address_for_webhook(self):
        self.test_resource.properties.data['type'] = self.test_resource.WEBHOOK
        self.test_resource.properties.data['address'] = 'https://'
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Address "https://" doesn\'t have '
                         'required network location', str(ex))

    def test_validate_prohibited_address_for_webhook(self):
        self.test_resource.properties.data['type'] = self.test_resource.WEBHOOK
        self.test_resource.properties.data['address'] = 'ftp://127.0.0.1'
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Address "ftp://127.0.0.1" doesn\'t satisfies '
                         'allowed schemes: http, https', str(ex))

    def test_validate_incorrect_address_for_email(self):
        self.test_resource.properties.data['type'] = self.test_resource.EMAIL
        self.test_resource.properties.data['address'] = 'abc#def.com'
        self.test_resource.properties.data.pop('period')
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Address "abc#def.com" doesn\'t satisfies allowed '
                         'format for "email" type of "type" property',
                         str(ex))

    def test_validate_invalid_address_parsing(self):
        self.test_resource.properties.data['type'] = self.test_resource.WEBHOOK
        self.test_resource.properties.data['address'] = "https://example.com]"
        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Address "https://example.com]" should have correct '
                         'format required by "webhook" type of "type" '
                         'property', str(ex))

    def test_resource_handle_create(self):
        mock_notification_create = self.test_client.notifications.create
        mock_resource = self._get_mock_resource()
        mock_notification_create.return_value = mock_resource

        # validate the properties
        self.assertEqual(
            'test-notification',
            self.test_resource.properties.get(
                notification.MonascaNotification.NAME))
        self.assertEqual(
            'webhook',
            self.test_resource.properties.get(
                notification.MonascaNotification.TYPE))
        self.assertEqual(
            'http://localhost:80/',
            self.test_resource.properties.get(
                notification.MonascaNotification.ADDRESS))
        self.assertEqual(
            60, self.test_resource.properties.get(
                notification.MonascaNotification.PERIOD))

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        args = dict(
            name='test-notification',
            type='webhook',
            address='http://localhost:80/',
            period=60
        )

        mock_notification_create.assert_called_once_with(**args)

        # validate physical resource id
        self.assertEqual(mock_resource['id'], self.test_resource.resource_id)

    def test_resource_handle_create_default_period(self):
        self.test_resource.properties.data.pop('period')
        mock_notification_create = self.test_client.notifications.create
        self.test_resource.handle_create()

        args = dict(
            name='test-notification',
            type='webhook',
            address='http://localhost:80/',
            period=60
        )
        mock_notification_create.assert_called_once_with(**args)

    def test_resource_handle_create_no_period(self):
        self.test_resource.properties.data.pop('period')
        self.test_resource.properties.data['type'] = 'email'
        self.test_resource.properties.data['address'] = 'abc@def.com'
        mock_notification_create = self.test_client.notifications.create
        self.test_resource.handle_create()

        args = dict(
            name='test-notification',
            type='email',
            address='abc@def.com'
        )
        mock_notification_create.assert_called_once_with(**args)

    def test_resource_handle_update(self):
        mock_notification_update = self.test_client.notifications.update
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {notification.MonascaNotification.ADDRESS:
                     'http://localhost:1234/',
                     notification.MonascaNotification.NAME: 'name-updated',
                     notification.MonascaNotification.TYPE: 'webhook',
                     notification.MonascaNotification.PERIOD: 0}

        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        args = dict(
            notification_id=self.test_resource.resource_id,
            name='name-updated',
            type='webhook',
            address='http://localhost:1234/',
            period=0
        )
        mock_notification_update.assert_called_once_with(**args)

    def test_resource_handle_update_default_period(self):
        mock_notification_update = self.test_client.notifications.update
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.test_resource.properties.data.pop('period')

        prop_diff = {notification.MonascaNotification.ADDRESS:
                     'http://localhost:1234/',
                     notification.MonascaNotification.NAME: 'name-updated',
                     notification.MonascaNotification.TYPE: 'webhook'}

        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        args = dict(
            notification_id=self.test_resource.resource_id,
            name='name-updated',
            type='webhook',
            address='http://localhost:1234/',
            period=60
        )
        mock_notification_update.assert_called_once_with(**args)

    def test_resource_handle_update_no_period(self):
        mock_notification_update = self.test_client.notifications.update
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.test_resource.properties.data.pop('period')

        prop_diff = {notification.MonascaNotification.ADDRESS:
                     'abc@def.com',
                     notification.MonascaNotification.NAME: 'name-updated',
                     notification.MonascaNotification.TYPE: 'email'}

        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        args = dict(
            notification_id=self.test_resource.resource_id,
            name='name-updated',
            type='email',
            address='abc@def.com'
        )
        mock_notification_update.assert_called_once_with(**args)

    def test_resource_handle_delete(self):
        mock_notification_delete = self.test_client.notifications.delete
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_notification_delete.return_value = None

        self.assertIsNone(self.test_resource.handle_delete())
        mock_notification_delete.assert_called_once_with(
            notification_id=self.test_resource.resource_id
        )

    def test_resource_handle_delete_resource_id_is_none(self):
        self.test_resource.resource_id = None
        self.assertIsNone(self.test_resource.handle_delete())

    def test_resource_handle_delete_not_found(self):
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_notification_delete = self.test_client.notifications.delete
        mock_notification_delete.side_effect = (
            client_plugin.monasca_exc.NotFound)

        self.assertIsNone(self.test_resource.handle_delete())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class NodeData(object):
    """Data about a node in the graph, to be passed along to other nodes."""

    __slots__ = ('primary_key', 'name', 'uuid',
                 '_reference_id', '_attributes',
                 'action', 'status')

    def __init__(self, primary_key, resource_name, uuid,
                 reference_id, attributes, action, status):
        """Initialise with data about the resource processed by the node.

        :param primary_key: the ID of the resource in the database
        :param name: the logical resource name
        :param uuid: the UUID of the resource
        :param reference_id: the value to be returned by get_resource
        :param attributes: dict of attributes values to be returned by get_attr
        :param action: the last resource action
        :param status: the status of the last action
        """
        self.primary_key = primary_key
        self.name = resource_name
        self.uuid = uuid
        self._reference_id = reference_id
        self._attributes = attributes
        self.action = action
        self.status = status

    def reference_id(self):
        """Return the reference ID of the resource.

        i.e. the result that the {get_resource: } intrinsic function should
        return for this resource.
        """
        return self._reference_id

    def attributes(self):
        """Return a dict of all available top-level attribute values."""
        attrs = {k: v
                 for k, v in self._attributes.items()
                 if isinstance(k, str)}
        for v in attrs.values():
            if isinstance(v, Exception):
                raise v
        return attrs

    def attribute(self, attr_name):
        """Return the specified attribute value."""
        val = self._attributes[attr_name]
        if isinstance(val, Exception):
            raise val
        return val

    def attribute_names(self):
        """Iterate over valid top-level attribute names."""
        for key in self._attributes:
            if isinstance(key, str):
                yield key
            else:
                yield key[0]

    def as_dict(self):
        """Return a dict representation of the data.

        This is the format that is serialised and stored in the database's
        SyncPoints.
        """
        for v in self._attributes.values():
            if isinstance(v, Exception):
                raise v

        return {
            'id': self.primary_key,
            'name': self.name,
            'reference_id': self.reference_id(),
            'attrs': dict(self._attributes),
            'status': self.status,
            'action': self.action,
            'uuid': self.uuid,
        }

    @classmethod
    def from_dict(cls, node_data):
        """Create a new NodeData object from deserialised data.

        This reads the format that is stored in the database, and is the
        inverse of as_dict().
        """
        if isinstance(node_data, cls):
            return node_data

        return cls(node_data.get('id'),
                   node_data.get('name'),
                   node_data.get('uuid'),
                   node_data.get('reference_id'),
                   node_data.get('attrs', {}),
                   node_data.get('action'),
                   node_data.get('status'))


def load_resources_data(data):
    """Return the data for all of the resources that meet at a SyncPoint.

    The input is the input_data dict from a SyncPoint received over RPC. The
    keys (which are ignored) are resource primary keys.

    The output is a dict of NodeData objects with the resource names as the
    keys.
    """
    nodes = (NodeData.from_dict(nd) for nd in data.values() if nd is not None)
    return {node.name: node for node in nodes}



# -*- coding: utf-8 -*-
# Copyright 2013 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import eventlet
from oslo_config import cfg
import oslo_messaging
from oslo_messaging.rpc import dispatcher
from osprofiler import profiler

from heat.common import context


TRANSPORT = None
NOTIFICATIONS_TRANSPORT = None
NOTIFIER = None


class RequestContextSerializer(oslo_messaging.Serializer):
    def __init__(self, base):
        self._base = base

    def serialize_entity(self, ctxt, entity):
        if not self._base:
            return entity
        return self._base.serialize_entity(ctxt, entity)

    def deserialize_entity(self, ctxt, entity):
        if not self._base:
            return entity
        return self._base.deserialize_entity(ctxt, entity)

    @staticmethod
    def serialize_context(ctxt):
        _context = ctxt.to_dict()
        prof = profiler.get()
        if prof:
            trace_info = {
                "hmac_key": prof.hmac_key,
                "base_id": prof.get_base_id(),
                "parent_id": prof.get_id()
            }
            _context.update({"trace_info": trace_info})
        return _context

    @staticmethod
    def deserialize_context(ctxt):
        trace_info = ctxt.pop("trace_info", None)
        if trace_info:
            profiler.init(**trace_info)
        return context.RequestContext.from_dict(ctxt)


def get_specific_transport(url, optional, exmods, is_for_notifications=False):
    try:
        if is_for_notifications:
            return oslo_messaging.get_notification_transport(
                cfg.CONF, url, allowed_remote_exmods=exmods)
        else:
            return oslo_messaging.get_rpc_transport(
                cfg.CONF, url, allowed_remote_exmods=exmods)
    except oslo_messaging.InvalidTransportURL as e:
        if not optional or e.url:
            # NOTE(sileht): oslo_messaging is configured but unloadable
            # so reraise the exception
            raise
        else:
            return None


def setup_transports(url, optional):
    global TRANSPORT, NOTIFICATIONS_TRANSPORT
    oslo_messaging.set_transport_defaults('heat')
    exmods = ['heat.common.exception']
    TRANSPORT = get_specific_transport(url, optional, exmods)
    NOTIFICATIONS_TRANSPORT = get_specific_transport(url, optional, exmods,
                                                     is_for_notifications=True)


def setup(url=None, optional=False):
    """Initialise the oslo_messaging layer."""
    global NOTIFIER

    if url and url.startswith("fake://"):
        # NOTE(sileht): oslo_messaging fake driver uses time.sleep
        # for task switch, so we need to monkey_patch it
        eventlet.monkey_patch(time=True)
    if not TRANSPORT or not NOTIFICATIONS_TRANSPORT:
        setup_transports(url, optional)
        # In the fake driver, make the dict of exchanges local to each exchange
        # manager, instead of using the shared class attribute. Doing otherwise
        # breaks the unit tests.
        if url and url.startswith("fake://"):
            TRANSPORT._driver._exchange_manager._exchanges = {}

    if not NOTIFIER and NOTIFICATIONS_TRANSPORT:
        serializer = RequestContextSerializer(
            oslo_messaging.JsonPayloadSerializer())
        NOTIFIER = oslo_messaging.Notifier(NOTIFICATIONS_TRANSPORT,
                                           serializer=serializer)


def cleanup():
    """Cleanup the oslo_messaging layer."""
    global TRANSPORT, NOTIFICATIONS_TRANSPORT, NOTIFIER
    if TRANSPORT:
        TRANSPORT.cleanup()
        NOTIFICATIONS_TRANSPORT.cleanup()
        TRANSPORT = NOTIFICATIONS_TRANSPORT = NOTIFIER = None


def get_rpc_server(target, endpoint):
    """Return a configured oslo_messaging rpc server."""
    serializer = RequestContextSerializer(
        oslo_messaging.JsonPayloadSerializer())
    access_policy = dispatcher.DefaultRPCAccessPolicy
    return oslo_messaging.get_rpc_server(TRANSPORT, target, [endpoint],
                                         executor='eventlet',
                                         serializer=serializer,
                                         access_policy=access_policy)


def get_rpc_client(**kwargs):
    """Return a configured oslo_messaging RPCClient."""
    target = oslo_messaging.Target(**kwargs)
    serializer = RequestContextSerializer(
        oslo_messaging.JsonPayloadSerializer())
    return oslo_messaging.get_rpc_client(
        TRANSPORT, target, serializer=serializer)


def get_notifier(publisher_id):
    """Return a configured oslo_messaging notifier."""
    return NOTIFIER.prepare(publisher_id=publisher_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from webob import exc

from heat.api.openstack.v1 import util
from heat.common.i18n import _
from heat.common import identifier
from heat.common import param_utils
from heat.common import serializers
from heat.common import wsgi
from heat.rpc import api as rpc_api
from heat.rpc import client as rpc_client


def format_resource(req, res, keys=None):
    keys = keys or []

    def include_key(k):
        return k in keys if keys else True

    def transform(key, value):
        if not include_key(key):
            return

        if key == rpc_api.RES_ID:
            identity = identifier.ResourceIdentifier(**value)
            links = [util.make_link(req, identity),
                     util.make_link(req, identity.stack(), 'stack')]

            nested_id = res.get(rpc_api.RES_NESTED_STACK_ID)
            if nested_id:
                nested_identity = identifier.HeatIdentifier(**nested_id)
                links.append(util.make_link(req, nested_identity, 'nested'))

            yield ('links', links)
        elif (key == rpc_api.RES_STACK_NAME or
              key == rpc_api.RES_STACK_ID or
              key == rpc_api.RES_ACTION or
              key == rpc_api.RES_NESTED_STACK_ID):
            return
        elif (key == rpc_api.RES_METADATA):
            return
        elif (key == rpc_api.RES_STATUS and rpc_api.RES_ACTION in res):
            # To avoid breaking API compatibility, we join RES_ACTION
            # and RES_STATUS, so the API format doesn't expose the
            # internal split of state into action/status
            yield (key, '_'.join((res[rpc_api.RES_ACTION], value)))
        elif (key == rpc_api.RES_NAME):
            yield ('logical_resource_id', value)
            yield (key, value)

        else:
            yield (key, value)

    return dict(itertools.chain.from_iterable(
        transform(k, v) for k, v in res.items()))


class ResourceController(object):
    """WSGI controller for Resources in Heat v1 API.

    Implements the API actions.
    """
    # Define request scope (must match what is in policy.yaml or policies in
    # code)
    REQUEST_SCOPE = 'resource'

    def __init__(self, options):
        self.options = options
        self.rpc_client = rpc_client.EngineClient()

    def _extract_to_param(self, req, rpc_param, extractor, default):
        key = rpc_param
        if key in req.params:
            try:
                return extractor(key, req.params[key])
            except ValueError as e:
                raise exc.HTTPBadRequest(str(e))
        else:
            return default

    @util.registered_identified_stack
    def index(self, req, identity):
        """Lists information for all resources."""

        param_types = {
            'type': util.PARAM_TYPE_MIXED,
            'status': util.PARAM_TYPE_MIXED,
            'name': util.PARAM_TYPE_MIXED,
            'action': util.PARAM_TYPE_MIXED,
            'id': util.PARAM_TYPE_MIXED,
            'physical_resource_id': util.PARAM_TYPE_MIXED,
        }

        invalid_keys = (set(req.params.keys()) -
                        set(list(param_types) + [rpc_api.PARAM_NESTED_DEPTH,
                                                 rpc_api.PARAM_WITH_DETAIL]))
        if invalid_keys:
            raise exc.HTTPBadRequest(_('Invalid filter parameters %s') %
                                     str(list(invalid_keys)))

        nested_depth = self._extract_to_param(req,
                                              rpc_api.PARAM_NESTED_DEPTH,
                                              param_utils.extract_int,
                                              default=0)
        with_detail = self._extract_to_param(req,
                                             rpc_api.PARAM_WITH_DETAIL,
                                             param_utils.extract_bool,
                                             default=False)

        params = util.get_allowed_params(req.params, param_types)

        res_list = self.rpc_client.list_stack_resources(req.context,
                                                        identity,
                                                        nested_depth,
                                                        with_detail,
                                                        filters=params)

        return {'resources': [format_resource(req, res) for res in res_list]}

    @util.registered_identified_stack
    def show(self, req, identity, resource_name):
        """Gets detailed information for a resource."""

        param_types = {'with_attr': util.PARAM_TYPE_MULTI}
        params = util.get_allowed_params(req.params, param_types)
        if 'with_attr' not in params:
            params['with_attr'] = None
        res = self.rpc_client.describe_stack_resource(req.context,
                                                      identity,
                                                      resource_name,
                                                      **params)

        return {'resource': format_resource(req, res)}

    @util.registered_identified_stack
    def metadata(self, req, identity, resource_name):
        """Gets metadata information for a resource."""

        res = self.rpc_client.describe_stack_resource(req.context,
                                                      identity,
                                                      resource_name)

        return {rpc_api.RES_METADATA: res[rpc_api.RES_METADATA]}

    @util.registered_identified_stack
    def signal(self, req, identity, resource_name, body=None):
        self.rpc_client.resource_signal(req.context,
                                        stack_identity=identity,
                                        resource_name=resource_name,
                                        details=body)

    @util.registered_identified_stack
    def mark_unhealthy(self, req, identity, resource_name, body):
        """Mark a resource as healthy or unhealthy."""
        data = dict()
        VALID_KEYS = (RES_UPDATE_MARK_UNHEALTHY, RES_UPDATE_STATUS_REASON) = (
            'mark_unhealthy', rpc_api.RES_STATUS_DATA)

        invalid_keys = set(body) - set(VALID_KEYS)
        if invalid_keys:
            raise exc.HTTPBadRequest(_("Invalid keys in resource "
                                       "mark unhealthy %s") % invalid_keys)

        if RES_UPDATE_MARK_UNHEALTHY not in body:
            raise exc.HTTPBadRequest(
                _("Missing mandatory (%s) key from mark unhealthy "
                  "request") % RES_UPDATE_MARK_UNHEALTHY)

        try:
            data[RES_UPDATE_MARK_UNHEALTHY] = param_utils.extract_bool(
                RES_UPDATE_MARK_UNHEALTHY,
                body[RES_UPDATE_MARK_UNHEALTHY])
        except ValueError as e:
            raise exc.HTTPBadRequest(str(e))

        data[RES_UPDATE_STATUS_REASON] = body.get(RES_UPDATE_STATUS_REASON, "")
        self.rpc_client.resource_mark_unhealthy(req.context,
                                                stack_identity=identity,
                                                resource_name=resource_name,
                                                **data)


def create_resource(options):
    """Resources resource factory method."""
    deserializer = wsgi.JSONRequestDeserializer()
    serializer = serializers.JSONResponseSerializer()
    return wsgi.Resource(ResourceController(options), deserializer, serializer)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class LifecyclePlugin(object):
    """Base class for pre-op and post-op work on a stack.

    Implementations should extend this class and override the methods.
    """
    def do_pre_op(self, cnxt, stack, current_stack=None, action=None):
        """Method to be run by heat before stack operations."""
        pass

    def do_post_op(self, cnxt, stack, current_stack=None, action=None,
                   is_stack_failure=False):
        """Method to be run by heat after stack operations, including failures.

        On failure to execute all the registered pre_ops, this method will be
        called if and only if the corresponding pre_op was successfully called.
        On failures of the actual stack operation, this method will
        be called if all the pre operations were successfully called.
        """
        pass

    def get_ordinal(self):
        """Get the sort order for pre and post operation execution.

        The values returned by get_ordinal are used to create a partial order
        for pre and post operation method invocations. The default ordinal
        value of 100 may be overridden.
        If class1inst.ordinal() < class2inst.ordinal(), then the method on
        class1inst will be executed before the method on class2inst.
        If class1inst.ordinal() > class2inst.ordinal(), then the method on
        class1inst will be executed after the method on class2inst.
        If class1inst.ordinal() == class2inst.ordinal(), then the order of
        method invocation is indeterminate.
        """
        return 100



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.tests.convergence.framework import engine_wrapper
from heat.tests.convergence.framework import event_loop as event_loop_module
from heat.tests.convergence.framework import worker_wrapper


engine = None
worker = None
event_loop = None


class Processes(object):

    def __init__(self):
        global engine
        global worker
        global event_loop

        worker = worker_wrapper.Worker()
        engine = engine_wrapper.Engine(worker)

        event_loop = event_loop_module.EventLoop(engine, worker)

        self.engine = engine
        self.worker = worker
        self.event_loop = event_loop

    def clear(self):
        self.engine.clear()
        self.worker.clear()


Processes()



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""SyncPoint object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.db import api as db_api
from heat.objects import base as heat_base
from heat.objects import fields as heat_fields


class SyncPoint(
        heat_base.HeatObject,
        base.VersionedObjectDictCompat,
        base.ComparableVersionedObject,
):

    fields = {
        'entity_id': fields.StringField(),
        'traversal_id': fields.StringField(),
        'is_update': fields.BooleanField(),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'atomic_key': fields.IntegerField(),
        'stack_id': fields.StringField(),
        'input_data': heat_fields.JsonField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, sdata, db_sdata):
        if db_sdata is None:
            return None
        for field in sdata.fields:
            sdata[field] = db_sdata[field]
        sdata._context = context
        sdata.obj_reset_changes()
        return sdata

    @classmethod
    def get_by_key(cls,
                   context,
                   entity_id,
                   traversal_id,
                   is_update):
        sync_point_db = db_api.sync_point_get(context,
                                              entity_id,
                                              traversal_id,
                                              is_update)
        return cls._from_db_object(context, cls(), sync_point_db)

    @classmethod
    def create(cls, context, values):
        sync_point_db = db_api.sync_point_create(context, values)
        return cls._from_db_object(context, cls(), sync_point_db)

    @classmethod
    def update_input_data(cls,
                          context,
                          entity_id,
                          traversal_id,
                          is_update,
                          atomic_key,
                          input_data):
        return db_api.sync_point_update_input_data(
            context,
            entity_id,
            traversal_id,
            is_update,
            atomic_key,
            input_data)

    @classmethod
    def delete_all_by_stack_and_traversal(cls,
                                          context,
                                          stack_id,
                                          traversal_id):
        return db_api.sync_point_delete_all_by_stack_and_traversal(
            context,
            stack_id,
            traversal_id)



#
# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1 import exceptions as keystone_exceptions
from keystoneauth1 import session
from webob import exc

from heat.common import config
from heat.common import context


class KeystonePasswordAuthProtocol(object):
    """Middleware uses username and password to authenticate against Keystone.

    Alternative authentication middleware that uses username and password
    to authenticate against Keystone instead of validating existing auth token.
    The benefit being that you no longer require admin/service token to
    authenticate users.
    """

    def __init__(self, app, conf):
        self.app = app
        self.conf = conf
        self.session = session.Session(**config.get_ssl_options('keystone'))

    def __call__(self, env, start_response):
        """Authenticate incoming request."""
        username = env.get('HTTP_X_AUTH_USER')
        password = env.get('HTTP_X_AUTH_KEY')
        # Determine project id from path.
        project_id = env.get('PATH_INFO').split('/')[1]
        auth_url = env.get('HTTP_X_AUTH_URL')
        user_domain_id = env.get('HTTP_X_USER_DOMAIN_ID')
        if not project_id:
            return self._reject_request(env, start_response, auth_url)
        try:
            ctx = context.RequestContext(username=username,
                                         password=password,
                                         project_id=project_id,
                                         auth_url=auth_url,
                                         user_domain_id=user_domain_id,
                                         is_admin=False)
            auth_ref = ctx.auth_plugin.get_access(self.session)
        except (keystone_exceptions.Unauthorized,
                keystone_exceptions.Forbidden,
                keystone_exceptions.NotFound,
                keystone_exceptions.AuthorizationFailure):
            return self._reject_request(env, start_response, auth_url)
        env.update(self._build_user_headers(auth_ref))

        return self.app(env, start_response)

    def _reject_request(self, env, start_response, auth_url):
        """Redirect client to auth server."""
        headers = [('WWW-Authenticate', "Keystone uri='%s'" % auth_url)]
        resp = exc.HTTPUnauthorized('Authentication required', headers)
        return resp(env, start_response)

    def _build_user_headers(self, token_info):
        """Build headers that represent authenticated user from auth token."""

        if token_info.version == 'v3':
            project_id = token_info.project_id
            project_name = token_info.project_name
        else:
            project_id = token_info.tenant_id
            project_name = token_info.tenant_name

        user_id = token_info.user_id
        user_name = token_info.username
        roles = ','.join(
            [role for role in token_info.role_names])
        service_catalog = token_info.service_catalog
        auth_token = token_info.auth_token
        user_domain_id = token_info.user_domain_id

        headers = {
            'HTTP_X_IDENTITY_STATUS': 'Confirmed',
            'HTTP_X_PROJECT_ID': project_id,
            'HTTP_X_PROJECT_NAME': project_name,
            'HTTP_X_USER_ID': user_id,
            'HTTP_X_USER_NAME': user_name,
            'HTTP_X_ROLES': roles,
            'HTTP_X_SERVICE_CATALOG': service_catalog,
            'HTTP_X_AUTH_TOKEN': auth_token,
            'HTTP_X_USER_DOMAIN_ID': user_domain_id,
        }

        return headers


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return KeystonePasswordAuthProtocol(app, conf)
    return auth_filter



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from heat.policies import base

DEPRECATED_REASON = """
The service API now supports system scope and default roles.
"""

POLICY_ROOT = 'service:%s'

deprecated_index = policy.DeprecatedRule(
    name=POLICY_ROOT % 'index',
    check_str=base.RULE_CONTEXT_IS_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

service_policies = [
    policy.RuleDefault(
        name=POLICY_ROOT % 'index',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        deprecated_rule=deprecated_index
    )
]


def list_rules():
    return service_policies



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.engine import resource
from heat.engine import support


class OctaviaBase(resource.Resource):

    default_client_name = 'octavia'

    support_status = support.SupportStatus(version='10.0.0')

    def _check_status(self, expected_status='ACTIVE'):
        res = self._show_resource()
        status = res['provisioning_status']
        if status == 'ERROR':
            raise exception.ResourceInError(resource_status=status)
        return status == expected_status

    def _check_deleted(self):
        with self.client_plugin().ignore_not_found:
            return self._check_status('DELETED')
        return True

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        attributes = self._show_resource()
        return attributes[name]

    def handle_create(self):
        return self._prepare_args(self.properties)

    def check_create_complete(self, properties):
        if self.resource_id is None:
            try:
                res = self._resource_create(properties)
                self.resource_id_set(res['id'])
            except Exception as ex:
                if self.client_plugin().is_conflict(ex):
                    return False
                raise

        return self._check_status()

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._update_called = False
        return prop_diff

    def check_update_complete(self, prop_diff):
        if not prop_diff:
            return True

        if not self._update_called:
            try:
                self._resource_update(prop_diff)
                self._update_called = True
            except Exception as ex:
                if self.client_plugin().is_conflict(ex):
                    return False
                raise

        return self._check_status()

    def handle_delete(self):
        self._delete_called = False

    def check_delete_complete(self, data):
        if self.resource_id is None:
            return True

        if not self._delete_called:
            try:
                self._resource_delete()
                self._delete_called = True
            except Exception as ex:
                if self.client_plugin().is_conflict(ex):
                    return self._check_status('DELETED')
                elif self.client_plugin().is_not_found(ex):
                    return True
                raise

        return self._check_deleted()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json

from heatclient import exc
import yaml

from heat_integrationtests.functional import functional_base


class ResourceGroupTest(functional_base.FunctionalTestsBase):
    template = '''
heat_template_version: 2013-05-23
resources:
  random_group:
    type: OS::Heat::ResourceGroup
    properties:
      count: 0
      resource_def:
        type: My::RandomString
        properties:
          length: 30
          salt: initial
outputs:
  random1:
    value: {get_attr: [random_group, resource.0.value]}
  random2:
    value: {get_attr: [random_group, resource.1.value]}
  all_values:
    value: {get_attr: [random_group, value]}
'''

    def test_resource_group_zero_novalidate(self):
        # Nested resources should be validated only when size > 0
        # This allows features to be disabled via size=0 without
        # triggering validation of nested resource custom constraints
        # e.g images etc in the nested schema.
        nested_template_fail = '''
heat_template_version: 2013-05-23
parameters:
  length:
    type: string
    default: 50
  salt:
    type: string
    default: initial
resources:
  random:
    type: OS::Heat::RandomString
    properties:
      length: BAD
'''

        files = {'provider.yaml': nested_template_fail}
        env = {'resource_registry':
               {'My::RandomString': 'provider.yaml'}}
        stack_identifier = self.stack_create(
            template=self.template,
            files=files,
            environment=env
        )

        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))

        # Check we created an empty nested stack
        nested_identifier = self.group_nested_identifier(stack_identifier,
                                                         'random_group')
        self.assertEqual({}, self.list_resources(nested_identifier))

        # Prove validation works for non-zero create/update
        template_two_nested = self.template.replace("count: 0", "count: 2")
        expected_err = ("resources.random_group<nested_stack>.resources."
                        "0<provider.yaml>.resources.random: : "
                        "Value 'BAD' is not an integer")
        ex = self.assertRaises(exc.HTTPBadRequest, self.update_stack,
                               stack_identifier, template_two_nested,
                               environment=env, files=files)
        self.assertIn(expected_err, str(ex))

        ex = self.assertRaises(exc.HTTPBadRequest, self.stack_create,
                               template=template_two_nested,
                               environment=env, files=files)
        self.assertIn(expected_err, str(ex))

    def _validate_resources(self, stack_identifier, expected_count):
        resources = self.list_group_resources(stack_identifier,
                                              'random_group')
        self.assertEqual(expected_count, len(resources))
        expected_resources = dict(
            (str(idx), 'My::RandomString')
            for idx in range(expected_count))

        self.assertEqual(expected_resources, resources)

    def test_create(self):
        def validate_output(stack, output_key, length):
            output_value = self._stack_output(stack, output_key)
            self.assertEqual(length, len(output_value))
            return output_value
        # verify that the resources in resource group are identically
        # configured, resource names and outputs are appropriate.
        env = {'resource_registry':
               {'My::RandomString': 'OS::Heat::RandomString'}}
        create_template = self.template.replace("count: 0", "count: 2")
        stack_identifier = self.stack_create(template=create_template,
                                             environment=env)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))

        # validate count, type and name of resources in a resource group.
        self._validate_resources(stack_identifier, 2)

        # validate outputs
        stack = self.client.stacks.get(stack_identifier)
        outputs = []
        outputs.append(validate_output(stack, 'random1', 30))
        outputs.append(validate_output(stack, 'random2', 30))
        self.assertEqual(outputs, self._stack_output(stack, 'all_values'))

    def test_create_nested_groups_with_timeout(self):
        parent_template = '''
heat_template_version: rocky
resources:
  parent_group:
    type: OS::Heat::ResourceGroup
    update_policy:
      batch_create: { max_batch_size: 1, pause_time: 1 }
    properties:
      count: 2
      resource_def:
        type: child.yaml
'''
        child_template = '''
heat_template_version: rocky
resources:
  child_group:
    type: OS::Heat::ResourceGroup
    update_policy:
      batch_create: { max_batch_size: 1, pause_time: 1 }
    properties:
      count: 2
      resource_def:
        type: value.yaml
'''
        value_template = '''
heat_template_version: rocky
resources:
  value:
    type: OS::Heat::Value
    properties:
      type: string
      value: 'test'
'''
        files = {
            'child.yaml': child_template,
            'value.yaml': value_template,
        }
        stack_identifier = self.stack_create(
            template=parent_template,
            files=files,
            timeout=10,  # in minutes
        )

        resources = self.client.resources.list(
            stack_identifier, nested_depth=2, with_detail=True)
        timeouts = set()
        for res in resources:
            if res.resource_type == "OS::Heat::ResourceGroup":
                nested_stack = self.client.stacks.get(res.physical_resource_id)
                timeouts.add(nested_stack.timeout_mins)

        self.assertEqual({10}, timeouts)

    def test_update_increase_decrease_count(self):
        # create stack with resource group count 2
        env = {'resource_registry':
               {'My::RandomString': 'OS::Heat::RandomString'}}
        create_template = self.template.replace("count: 0", "count: 2")
        stack_identifier = self.stack_create(template=create_template,
                                             environment=env)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))
        # verify that the resource group has 2 resources
        self._validate_resources(stack_identifier, 2)

        # increase the resource group count to 5
        update_template = self.template.replace("count: 0", "count: 5")
        self.update_stack(stack_identifier, update_template, environment=env)
        # verify that the resource group has 5 resources
        self._validate_resources(stack_identifier, 5)

        # decrease the resource group count to 3
        update_template = self.template.replace("count: 0", "count: 3")
        self.update_stack(stack_identifier, update_template, environment=env)
        # verify that the resource group has 3 resources
        self._validate_resources(stack_identifier, 3)

    def test_update_removal_policies(self):
        rp_template = '''
heat_template_version: 2014-10-16
resources:
  random_group:
    type: OS::Heat::ResourceGroup
    properties:
      count: 5
      removal_policies: []
      resource_def:
        type: OS::Heat::RandomString
'''

        # create stack with resource group, initial count 5
        stack_identifier = self.stack_create(template=rp_template)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))
        group_resources = self.list_group_resources(stack_identifier,
                                                    'random_group')
        expected_resources = {'0': 'OS::Heat::RandomString',
                              '1': 'OS::Heat::RandomString',
                              '2': 'OS::Heat::RandomString',
                              '3': 'OS::Heat::RandomString',
                              '4': 'OS::Heat::RandomString'}
        self.assertEqual(expected_resources, group_resources)

        # Remove three, specifying the middle resources to be removed
        update_template = rp_template.replace(
            'removal_policies: []',
            'removal_policies: [{resource_list: [\'1\', \'2\', \'3\']}]')
        self.update_stack(stack_identifier, update_template)
        group_resources = self.list_group_resources(stack_identifier,
                                                    'random_group')
        expected_resources = {'0': 'OS::Heat::RandomString',
                              '4': 'OS::Heat::RandomString',
                              '5': 'OS::Heat::RandomString',
                              '6': 'OS::Heat::RandomString',
                              '7': 'OS::Heat::RandomString'}
        self.assertEqual(expected_resources, group_resources)

    def test_props_update(self):
        """Test update of resource_def properties behaves as expected."""

        env = {'resource_registry':
               {'My::RandomString': 'OS::Heat::RandomString'}}
        template_one = self.template.replace("count: 0", "count: 1")
        stack_identifier = self.stack_create(template=template_one,
                                             environment=env)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))

        initial_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual({'0': 'My::RandomString'},
                         self.list_resources(initial_nested_ident))
        # get the resource id
        res = self.client.resources.get(initial_nested_ident, '0')
        initial_res_id = res.physical_resource_id

        # change the salt (this should replace the RandomString but
        # not the nested stack or resource group.
        template_salt = template_one.replace("salt: initial", "salt: more")
        self.update_stack(stack_identifier, template_salt, environment=env)
        updated_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual(initial_nested_ident, updated_nested_ident)

        # compare the resource id, we expect a change.
        res = self.client.resources.get(updated_nested_ident, '0')
        updated_res_id = res.physical_resource_id
        self.assertNotEqual(initial_res_id, updated_res_id)

    def test_update_nochange(self):
        """Test update with no properties change."""

        env = {'resource_registry':
               {'My::RandomString': 'OS::Heat::RandomString'}}
        template_one = self.template.replace("count: 0", "count: 2")
        stack_identifier = self.stack_create(template=template_one,
                                             environment=env)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))

        initial_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual({'0': 'My::RandomString', '1': 'My::RandomString'},
                         self.list_resources(initial_nested_ident))
        # get the output
        stack0 = self.client.stacks.get(stack_identifier)
        initial_rand = self._stack_output(stack0, 'random1')

        template_copy = copy.deepcopy(template_one)
        self.update_stack(stack_identifier, template_copy, environment=env)
        updated_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual(initial_nested_ident, updated_nested_ident)

        # compare the random number, we expect no change.
        stack1 = self.client.stacks.get(stack_identifier)
        updated_rand = self._stack_output(stack1, 'random1')
        self.assertEqual(initial_rand, updated_rand)

    def test_update_nochange_resource_needs_update(self):
        """Test update when the resource definition has changed.

        Test the scenario when the ResourceGroup update happens without
        any changed properties, this can happen if the definition of
        a contained provider resource changes (files map changes), then
        the group and underlying nested stack should end up updated.
        """

        random_templ1 = '''
heat_template_version: 2013-05-23
parameters:
  length:
    type: string
    default: not-used
  salt:
    type: string
    default: not-used
resources:
  random1:
    type: OS::Heat::RandomString
    properties:
      salt: initial
outputs:
  value:
    value: {get_attr: [random1, value]}
'''
        files1 = {'my_random.yaml': random_templ1}

        random_templ2 = random_templ1.replace('salt: initial',
                                              'salt: more')
        files2 = {'my_random.yaml': random_templ2}

        env = {'resource_registry':
               {'My::RandomString': 'my_random.yaml'}}

        template_one = self.template.replace("count: 0", "count: 2")
        stack_identifier = self.stack_create(template=template_one,
                                             environment=env,
                                             files=files1)
        self.assertEqual({'random_group': 'OS::Heat::ResourceGroup'},
                         self.list_resources(stack_identifier))
        self.assertEqual(files1, self.client.stacks.files(stack_identifier))

        initial_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual({'0': 'My::RandomString', '1': 'My::RandomString'},
                         self.list_resources(initial_nested_ident))
        # get the output
        stack0 = self.client.stacks.get(stack_identifier)
        initial_rand = self._stack_output(stack0, 'random1')

        # change the environment so we use a different TemplateResource.
        # note "files2".
        self.update_stack(stack_identifier, template_one,
                          environment=env, files=files2)
        updated_nested_ident = self.group_nested_identifier(stack_identifier,
                                                            'random_group')
        self.assertEqual(initial_nested_ident, updated_nested_ident)
        self.assertEqual(files2, self.client.stacks.files(stack_identifier))

        # compare the output, we expect a change.
        stack1 = self.client.stacks.get(stack_identifier)
        updated_rand = self._stack_output(stack1, 'random1')
        self.assertNotEqual(initial_rand, updated_rand)

    def test_validation(self):
        resource_group = '''
heat_template_version: 2016-10-14

parameters:
  the_count:
    type: number

resources:

  the_group:
    type: OS::Heat::ResourceGroup
    properties:
      count: {get_param: the_count}
      resource_def:
        type: OS::Heat::RandomString
'''
        ret = self.client.stacks.validate(template=resource_group)
        expected = {'Description': 'No description',
                    'Environment': {'event_sinks': [],
                                    'parameter_defaults': {},
                                    'parameters': {},
                                    'resource_registry': {'resources': {}}},
                    'Parameters': {
                        'the_count': {'Description': '',
                                      'Label': 'the_count',
                                      'NoEcho': 'false',
                                      'Type': 'Number'}}}

        self.assertEqual(expected, ret)


class ResourceGroupTestNullParams(functional_base.FunctionalTestsBase):
    template = '''
heat_template_version: 2013-05-23
parameters:
  param:
    type: empty
resources:
  random_group:
    type: OS::Heat::ResourceGroup
    properties:
      count: 1
      resource_def:
        type: My::RandomString
        properties:
          param: {get_param: param}
outputs:
  val:
    value: {get_attr: [random_group, val]}
'''

    nested_template_file = '''
heat_template_version: 2013-05-23
parameters:
  param:
    type: empty
outputs:
  val:
    value: {get_param: param}
'''

    scenarios = [
        ('string_empty', dict(
            param='',
            p_type='string',
        )),
        ('boolean_false', dict(
            param=False,
            p_type='boolean',
        )),
        ('number_zero', dict(
            param=0,
            p_type='number',
        )),
        ('comma_delimited_list', dict(
            param=[],
            p_type='comma_delimited_list',
        )),
        ('json_empty', dict(
            param={},
            p_type='json',
        )),
    ]

    def test_create_pass_zero_parameter(self):
        templ = self.template.replace('type: empty',
                                      'type: %s' % self.p_type)
        n_t_f = self.nested_template_file.replace('type: empty',
                                                  'type: %s' % self.p_type)
        files = {'provider.yaml': n_t_f}
        env = {'resource_registry':
               {'My::RandomString': 'provider.yaml'}}
        stack_identifier = self.stack_create(
            template=templ,
            files=files,
            environment=env,
            parameters={'param': self.param}
        )
        stack = self.client.stacks.get(stack_identifier)
        self.assertEqual(self.param, self._stack_output(stack, 'val')[0])


class ResourceGroupAdoptTest(functional_base.FunctionalTestsBase):
    """Prove that we can do resource group adopt."""

    main_template = '''
heat_template_version: "2013-05-23"
resources:
  group1:
    type: OS::Heat::ResourceGroup
    properties:
      count: 2
      resource_def:
        type: OS::Heat::RandomString
outputs:
  test0:
    value: {get_attr: [group1, resource.0.value]}
  test1:
    value: {get_attr: [group1, resource.1.value]}
'''

    def _yaml_to_json(self, yaml_templ):
        return yaml.safe_load(yaml_templ)

    def test_adopt(self):
        data = {
            "resources": {
                "group1": {
                    "status": "COMPLETE",
                    "name": "group1",
                    "resource_data": {},
                    "metadata": {},
                    "resource_id": "test-group1-id",
                    "action": "CREATE",
                    "type": "OS::Heat::ResourceGroup",
                    "resources": {
                        "0": {
                            "status": "COMPLETE",
                            "name": "0",
                            "resource_data": {"value": "goopie"},
                            "resource_id": "ID-0",
                            "action": "CREATE",
                            "type": "OS::Heat::RandomString",
                            "metadata": {}
                        },
                        "1": {
                            "status": "COMPLETE",
                            "name": "1",
                            "resource_data": {"value": "different"},
                            "resource_id": "ID-1",
                            "action": "CREATE",
                            "type": "OS::Heat::RandomString",
                            "metadata": {}
                        }
                    }
                }
            },
            "environment": {"parameters": {}},
            "template": yaml.safe_load(self.main_template)
        }
        stack_identifier = self.stack_adopt(
            adopt_data=json.dumps(data))

        self.assert_resource_is_a_stack(stack_identifier, 'group1')
        stack = self.client.stacks.get(stack_identifier)
        self.assertEqual('goopie', self._stack_output(stack, 'test0'))
        self.assertEqual('different', self._stack_output(stack, 'test1'))


class ResourceGroupErrorResourceTest(functional_base.FunctionalTestsBase):
    template = '''
heat_template_version: "2013-05-23"
resources:
  group1:
    type: OS::Heat::ResourceGroup
    properties:
      count: 2
      resource_def:
        type: fail.yaml
'''
    nested_templ = '''
heat_template_version: "2013-05-23"
resources:
  oops:
    type: OS::Heat::TestResource
    properties:
      fail: true
      wait_secs: 2
'''

    def test_fail(self):
        stack_identifier = self.stack_create(
            template=self.template,
            files={'fail.yaml': self.nested_templ},
            expected_status='CREATE_FAILED',
            enable_cleanup=False)
        stack = self.client.stacks.get(stack_identifier)

        self.assertEqual('CREATE_FAILED', stack.stack_status)
        self.client.stacks.delete(stack_identifier)
        self._wait_for_stack_status(
            stack_identifier, 'DELETE_COMPLETE',
            success_on_not_found=True)


class ResourceGroupUpdatePolicyTest(functional_base.FunctionalTestsBase):

    template = '''
heat_template_version: '2015-04-30'
resources:
  random_group:
    type: OS::Heat::ResourceGroup
    update_policy:
      rolling_update:
        min_in_service: 1
        max_batch_size: 2
        pause_time: 1
    properties:
      count: 10
      resource_def:
        type: OS::Heat::TestResource
        properties:
          value: initial
          update_replace: False
'''

    def update_resource_group(self, update_template,
                              updated, created, deleted):
        stack_identifier = self.stack_create(template=self.template)
        group_resources = self.list_group_resources(stack_identifier,
                                                    'random_group',
                                                    minimal=False)

        init_names = [res.physical_resource_id for res in group_resources]

        self.update_stack(stack_identifier, update_template)
        group_resources = self.list_group_resources(stack_identifier,
                                                    'random_group',
                                                    minimal=False)

        updt_names = [res.physical_resource_id for res in group_resources]

        matched_names = set(updt_names) & set(init_names)

        self.assertEqual(updated, len(matched_names))

        self.assertEqual(created, len(set(updt_names) - set(init_names)))

        self.assertEqual(deleted, len(set(init_names) - set(updt_names)))

    def test_resource_group_update(self):
        """Test rolling update with no conflict.

        Simple rolling update with no conflict in batch size
        and minimum instances in service.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '1'
        policy['max_batch_size'] = '3'
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=10,
                                   created=0,
                                   deleted=0)

    def test_resource_group_update_replace(self):
        """Test rolling update(replace)with no conflict.

        Simple rolling update replace with no conflict in batch size
        and minimum instances in service.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '1'
        policy['max_batch_size'] = '3'
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'
        res_def['properties']['update_replace'] = True

        self.update_resource_group(updt_template,
                                   updated=0,
                                   created=10,
                                   deleted=10)

    def test_resource_group_update_replace_template_changed(self):
        """Test rolling update(replace)with child template path changed.

        Simple rolling update replace with child template path changed.
        """

        nested_templ = '''
heat_template_version: "2013-05-23"
resources:
  oops:
    type: OS::Heat::TestResource
'''

        create_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = create_template['resources']['random_group']
        grp['properties']['resource_def'] = {'type': '/opt/provider.yaml'}
        files = {'/opt/provider.yaml': nested_templ}

        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '1'
        policy['max_batch_size'] = '3'
        stack_identifier = self.stack_create(template=create_template,
                                             files=files)
        update_template = create_template.copy()
        grp = update_template['resources']['random_group']
        grp['properties']['resource_def'] = {'type': '/opt1/provider.yaml'}
        files = {'/opt1/provider.yaml': nested_templ}

        self.update_stack(stack_identifier, update_template, files=files)

    def test_resource_group_update_scaledown(self):
        """Test rolling update with scaledown.

        Simple rolling update with reduced size.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '1'
        policy['max_batch_size'] = '3'
        grp['properties']['count'] = 6
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=6,
                                   created=0,
                                   deleted=4)

    def test_resource_group_update_scaleup(self):
        """Test rolling update with scaleup.

        Simple rolling update with increased size.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '1'
        policy['max_batch_size'] = '3'
        grp['properties']['count'] = 12
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=10,
                                   created=2,
                                   deleted=0)

    def test_resource_group_update_adjusted(self):
        """Test rolling update with enough available resources

        Update  with capacity adjustment with enough resources.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '8'
        policy['max_batch_size'] = '4'
        grp['properties']['count'] = 6
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=6,
                                   created=0,
                                   deleted=4)

    def test_resource_group_update_with_adjusted_capacity(self):
        """Test rolling update with capacity adjustment.

        Rolling update with capacity adjustment due to conflict in
        batch size and minimum instances in service.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '8'
        policy['max_batch_size'] = '4'
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=10,
                                   created=0,
                                   deleted=0)

    def test_resource_group_update_huge_batch_size(self):
        """Test rolling update with huge batch size.

        Rolling Update with a huge batch size(more than
        current size).
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '0'
        policy['max_batch_size'] = '20'
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'
        self.update_resource_group(updt_template,
                                   updated=10,
                                   created=0,
                                   deleted=0)

    def test_resource_group_update_huge_min_in_service(self):
        """Test rolling update with huge minimum capacity.

        Rolling Update with a huge number of minimum instances
        in service.
        """
        updt_template = yaml.safe_load(copy.deepcopy(self.template))
        grp = updt_template['resources']['random_group']
        policy = grp['update_policy']['rolling_update']
        policy['min_in_service'] = '20'
        policy['max_batch_size'] = '1'
        res_def = grp['properties']['resource_def']
        res_def['properties']['value'] = 'updated'

        self.update_resource_group(updt_template,
                                   updated=10,
                                   created=0,
                                   deleted=0)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.engine import support
from heat.tests import common


class SupportStatusTest(common.HeatTestCase):
    def test_valid_status(self):
        for sstatus in support.SUPPORT_STATUSES:
            previous = support.SupportStatus(version='test_version')
            status = support.SupportStatus(
                status=sstatus,
                message='test_message',
                version='test_version',
                previous_status=previous,
            )
            self.assertEqual(sstatus, status.status)
            self.assertEqual('test_message', status.message)
            self.assertEqual('test_version', status.version)
            self.assertEqual(previous, status.previous_status)
            self.assertEqual({
                'status': sstatus,
                'message': 'test_message',
                'version': 'test_version',
                'previous_status': {'status': 'SUPPORTED',
                                    'message': None,
                                    'version': 'test_version',
                                    'previous_status': None},
            }, status.to_dict())

    def test_invalid_status(self):
        status = support.SupportStatus(
            status='RANDOM',
            message='test_message',
            version='test_version',
            previous_status=support.SupportStatus()
        )
        self.assertEqual(support.UNKNOWN, status.status)
        self.assertEqual('Specified status is invalid, defaulting to UNKNOWN',
                         status.message)
        self.assertIsNone(status.version)
        self.assertIsNone(status.previous_status)
        self.assertEqual({
            'status': 'UNKNOWN',
            'message': 'Specified status is invalid, defaulting to UNKNOWN',
            'version': None,
            'previous_status': None,
        }, status.to_dict())

    def test_previous_status(self):
        sstatus = support.SupportStatus(
            status=support.DEPRECATED,
            version='5.0.0',
            previous_status=support.SupportStatus(
                status=support.SUPPORTED,
                version='2015.1'
            )
        )

        self.assertEqual(support.DEPRECATED, sstatus.status)
        self.assertEqual('5.0.0', sstatus.version)
        self.assertEqual(support.SUPPORTED, sstatus.previous_status.status)
        self.assertEqual('2015.1', sstatus.previous_status.version)

        self.assertEqual({'status': 'DEPRECATED',
                          'version': '5.0.0',
                          'message': None,
                          'previous_status': {'status': 'SUPPORTED',
                                              'version': '2015.1',
                                              'message': None,
                                              'previous_status': None}},
                         sstatus.to_dict())

    def test_invalid_previous_status(self):
        ex = self.assertRaises(ValueError,
                               support.SupportStatus, previous_status='YARRR')
        self.assertEqual('previous_status must be SupportStatus '
                         'instead of %s' % str, str(ex))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat_integrationtests.functional import functional_base

test_template = '''
heat_template_version: 2014-10-16

resources:
  signal_handle:
    type: "OS::Heat::SwiftSignalHandle"

outputs:
  signal_curl:
    value: { get_attr: ['signal_handle', 'curl_cli'] }
    description: Swift signal cURL

  signal_url:
    value: { get_attr: ['signal_handle', 'endpoint'] }
    description: Swift signal URL
'''


class SwiftSignalHandleUpdateTest(functional_base.FunctionalTestsBase):

    def test_stack_update_same_template_replace_no_url(self):
        if not self.is_service_available('object-store'):
            self.skipTest('object-store service not available, skipping')
        stack_identifier = self.stack_create(template=test_template)
        stack = self.client.stacks.get(stack_identifier)
        orig_url = self._stack_output(stack, 'signal_url')
        orig_curl = self._stack_output(stack, 'signal_curl')
        self.update_stack(stack_identifier, test_template)
        stack = self.client.stacks.get(stack_identifier)
        self.assertEqual(orig_url, self._stack_output(stack, 'signal_url'))
        self.assertEqual(orig_curl, self._stack_output(stack, 'signal_curl'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base

test_template_one_resource = {
    'heat_template_version': '2013-05-23',
    'description': 'Test template to create one instance.',
    'resources': {
        'test1': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1',
                'fail': False,
                'update_replace': False,
                'wait_secs': 0
            }
        }
    }
}

test_template_two_resource = {
    'heat_template_version': '2013-05-23',
    'description': 'Test template to create two instance.',
    'resources': {
        'test1': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1',
                'fail': False,
                'update_replace': False,
                'wait_secs': 0
            }
        },
        'test2': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1',
                'fail': False,
                'update_replace': False,
                'wait_secs': 0
            }
        }
    }
}


class UpdatePreviewBase(functional_base.FunctionalTestsBase):

    def assert_empty_sections(self, changes, empty_sections):
        for section in empty_sections:
            self.assertEqual([], changes[section])


class UpdatePreviewStackTest(UpdatePreviewBase):

    def test_add_resource(self):
        self.stack_identifier = self.stack_create(
            template=test_template_one_resource)
        result = self.preview_update_stack(self.stack_identifier,
                                           test_template_two_resource)
        changes = result['resource_changes']

        unchanged = changes['unchanged'][0]['resource_name']
        self.assertEqual('test1', unchanged)

        added = changes['added'][0]['resource_name']
        self.assertEqual('test2', added)

        self.assert_empty_sections(changes, ['updated', 'replaced', 'deleted'])

    def test_no_change(self):
        self.stack_identifier = self.stack_create(
            template=test_template_one_resource)
        result = self.preview_update_stack(self.stack_identifier,
                                           test_template_one_resource)
        changes = result['resource_changes']

        unchanged = changes['unchanged'][0]['resource_name']
        self.assertEqual('test1', unchanged)

        self.assert_empty_sections(
            changes, ['updated', 'replaced', 'deleted', 'added'])

    def test_update_resource(self):
        self.stack_identifier = self.stack_create(
            template=test_template_one_resource)
        test_template_updated_resource = {
            'heat_template_version': '2013-05-23',
            'description': 'Test template to create one instance.',
            'resources': {
                'test1': {
                    'type': 'OS::Heat::TestResource',
                    'properties': {
                        'value': 'Test1 foo',
                        'fail': False,
                        'update_replace': False,
                        'wait_secs': 0
                    }
                }
            }
        }

        result = self.preview_update_stack(self.stack_identifier,
                                           test_template_updated_resource)
        changes = result['resource_changes']

        updated = changes['updated'][0]['resource_name']
        self.assertEqual('test1', updated)

        self.assert_empty_sections(
            changes, ['added', 'unchanged', 'replaced', 'deleted'])

    def test_replaced_resource(self):
        self.stack_identifier = self.stack_create(
            template=test_template_one_resource)
        new_template = {
            'heat_template_version': '2013-05-23',
            'description': 'Test template to create one instance.',
            'resources': {
                'test1': {
                    'type': 'OS::Heat::TestResource',
                    'properties': {
                        'update_replace': True,
                    }
                }
            }
        }

        result = self.preview_update_stack(self.stack_identifier, new_template)
        changes = result['resource_changes']

        replaced = changes['replaced'][0]['resource_name']
        self.assertEqual('test1', replaced)

        self.assert_empty_sections(
            changes, ['added', 'unchanged', 'updated', 'deleted'])

    def test_delete_resource(self):
        self.stack_identifier = self.stack_create(
            template=test_template_two_resource)
        result = self.preview_update_stack(self.stack_identifier,
                                           test_template_one_resource)
        changes = result['resource_changes']

        unchanged = changes['unchanged'][0]['resource_name']
        self.assertEqual('test1', unchanged)

        deleted = changes['deleted'][0]['resource_name']
        self.assertEqual('test2', deleted)

        self.assert_empty_sections(changes, ['updated', 'replaced', 'added'])


class UpdatePreviewStackTestNested(UpdatePreviewBase):
    template_nested_parent = '''
heat_template_version: 2016-04-08
resources:
  nested1:
    type: nested1.yaml
'''

    template_nested1 = '''
heat_template_version: 2016-04-08
resources:
  nested2:
    type: nested2.yaml
'''

    template_nested2 = '''
heat_template_version: 2016-04-08
resources:
  random:
    type: OS::Heat::RandomString
'''

    template_nested2_2 = '''
heat_template_version: 2016-04-08
resources:
  random:
    type: OS::Heat::RandomString
  random2:
    type: OS::Heat::RandomString
'''

    def _get_by_resource_name(self, changes, name, action):
        filtered_l = [x for x in changes[action]
                      if x['resource_name'] == name]
        self.assertEqual(1, len(filtered_l))
        return filtered_l[0]

    def test_nested_resources_nochange(self):
        files = {'nested1.yaml': self.template_nested1,
                 'nested2.yaml': self.template_nested2}
        self.stack_identifier = self.stack_create(
            template=self.template_nested_parent, files=files)
        result = self.preview_update_stack(
            self.stack_identifier,
            template=self.template_nested_parent,
            files=files, show_nested=True)
        changes = result['resource_changes']

        # The nested random resource should be unchanged, but we always
        # update nested stacks even when there are no changes
        self.assertEqual(1, len(changes['unchanged']))
        self.assertEqual('random', changes['unchanged'][0]['resource_name'])
        self.assertEqual('nested2', changes['unchanged'][0]['parent_resource'])

        self.assertEqual(2, len(changes['updated']))
        u_nested1 = self._get_by_resource_name(changes, 'nested1', 'updated')
        self.assertNotIn('parent_resource', u_nested1)
        u_nested2 = self._get_by_resource_name(changes, 'nested2', 'updated')
        self.assertEqual('nested1', u_nested2['parent_resource'])

        self.assert_empty_sections(changes, ['replaced', 'deleted', 'added'])

    def test_nested_resources_add(self):
        files = {'nested1.yaml': self.template_nested1,
                 'nested2.yaml': self.template_nested2}
        self.stack_identifier = self.stack_create(
            template=self.template_nested_parent, files=files)
        files['nested2.yaml'] = self.template_nested2_2
        result = self.preview_update_stack(
            self.stack_identifier,
            template=self.template_nested_parent,
            files=files, show_nested=True)
        changes = result['resource_changes']

        # The nested random resource should be unchanged, but we always
        # update nested stacks even when there are no changes
        self.assertEqual(1, len(changes['unchanged']))
        self.assertEqual('random', changes['unchanged'][0]['resource_name'])
        self.assertEqual('nested2', changes['unchanged'][0]['parent_resource'])

        self.assertEqual(1, len(changes['added']))
        self.assertEqual('random2', changes['added'][0]['resource_name'])
        self.assertEqual('nested2', changes['added'][0]['parent_resource'])

        self.assert_empty_sections(changes, ['replaced', 'deleted'])

    def test_nested_resources_delete(self):
        files = {'nested1.yaml': self.template_nested1,
                 'nested2.yaml': self.template_nested2_2}
        self.stack_identifier = self.stack_create(
            template=self.template_nested_parent, files=files)
        files['nested2.yaml'] = self.template_nested2
        result = self.preview_update_stack(
            self.stack_identifier,
            template=self.template_nested_parent,
            files=files, show_nested=True)
        changes = result['resource_changes']

        # The nested random resource should be unchanged, but we always
        # update nested stacks even when there are no changes
        self.assertEqual(1, len(changes['unchanged']))
        self.assertEqual('random', changes['unchanged'][0]['resource_name'])
        self.assertEqual('nested2', changes['unchanged'][0]['parent_resource'])

        self.assertEqual(1, len(changes['deleted']))
        self.assertEqual('random2', changes['deleted'][0]['resource_name'])
        self.assertEqual('nested2', changes['deleted'][0]['parent_resource'])

        self.assert_empty_sections(changes, ['replaced', 'added'])

    def test_nested_resources_replace(self):
        files = {'nested1.yaml': self.template_nested1,
                 'nested2.yaml': self.template_nested2}
        self.stack_identifier = self.stack_create(
            template=self.template_nested_parent, files=files)
        parent_none = self.template_nested_parent.replace(
            'nested1.yaml', 'OS::Heat::None')
        result = self.preview_update_stack(
            self.stack_identifier,
            template=parent_none,
            show_nested=True)
        changes = result['resource_changes']

        # The nested random resource should be unchanged, but we always
        # update nested stacks even when there are no changes
        self.assertEqual(1, len(changes['replaced']))
        self.assertEqual('nested1', changes['replaced'][0]['resource_name'])

        self.assertEqual(2, len(changes['deleted']))
        d_random = self._get_by_resource_name(changes, 'random', 'deleted')
        self.assertEqual('nested2', d_random['parent_resource'])
        d_nested2 = self._get_by_resource_name(changes, 'nested2', 'deleted')
        self.assertEqual('nested1', d_nested2['parent_resource'])

        self.assert_empty_sections(changes, ['updated', 'unchanged', 'added'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.common import template_format
from heat.engine import properties
from heat.engine.resources.openstack.heat import software_config
from heat.engine import support
from heat.rpc import api as rpc_api


class CloudConfig(software_config.SoftwareConfig):
    """A configuration resource for representing cloud-init cloud-config.

    This resource allows cloud-config YAML to be defined and stored by the
    config API. Any intrinsic functions called in the config will be resolved
    before storing the result.

    This resource will generally be referenced by OS::Nova::Server user_data,
    or OS::Heat::MultipartMime parts config. Since cloud-config is boot-only
    configuration, any changes to the definition will result in the
    replacement of all servers which reference it.
    """

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        CLOUD_CONFIG
    ) = (
        'cloud_config'
    )

    properties_schema = {
        CLOUD_CONFIG: properties.Schema(
            properties.Schema.MAP,
            _('Map representing the cloud-config data structure which will '
              'be formatted as YAML.')
        )
    }

    def handle_create(self):
        cloud_config = template_format.yaml.dump(
            self.properties[self.CLOUD_CONFIG],
            Dumper=template_format.yaml_dumper)
        props = {
            rpc_api.SOFTWARE_CONFIG_NAME: self.physical_resource_name(),
            rpc_api.SOFTWARE_CONFIG_CONFIG: '#cloud-config\n%s' % cloud_config,
            rpc_api.SOFTWARE_CONFIG_GROUP: 'Heat::Ungrouped'
        }
        sc = self.rpc_client().create_software_config(self.context, **props)
        self.resource_id_set(sc[rpc_api.SOFTWARE_CONFIG_ID])


def resource_mapping():
    return {
        'OS::Heat::CloudConfig': CloudConfig,
    }
